# This file is part of the Hotwire Shell user interface.
#
# Copyright (C) 2007 Colin Walters <walters@verbum.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os,sys,platform,logging,getopt,re
import locale,threading,subprocess,time
import signal,tempfile,shutil,stat,pwd
import datetime,gettext,sha,commands,errno
import urllib
from StringIO import StringIO
import xml.dom.minidom

try:
    import sqlite3
except:
    from pysqlite2 import dbapi2 as sqlite3

import gtk,gobject,pango
import dbus,dbus.glib,dbus.service

try:
    import avahi
    avahi_available = True
except:
    avahi_available = False

from hotssh.hotvte.vteterm import VteTerminalWidget
from hotssh.hotvte.vtewindow import VteWindow
from hotssh.hotvte.vtewindow import VteApp
from hotssh.hotlib.logutil import log_except
from hotssh.hotlib.timesince import timesince
from hotssh.hotlib_ui.quickfind import QuickFindWindow
from hotssh.hotlib_ui.msgarea import MsgAreaController

_logger = logging.getLogger("hotssh.SshWindow")

_whitespace_re = re.compile('\s+')

def glibc_backtrace():
    from ctypes import cdll
    libc = cdll.LoadLibrary("libc.so.6")
    btcount = 100
    btdata = (c_void_p * btcount)()
    libc.backtrace(btdata, c_int(btcount))
    symbols = libc.backtrace_symbols(btdata, c_int(btcount))
    print symbols

def userhost_pair_to_string(user, host, port=None):
    host = hostport_pair_to_string(host, port)
    if user:
        return user + '@' + host
    return host

def hostport_pair_to_string(host, port=None):
    hostport = host
    if port is not None:
        hostport += (':%s' % (port,))
    return hostport

class HotSshAboutDialog(gtk.AboutDialog):
    def __init__(self):
        super(HotSshAboutDialog, self).__init__()
        dialog = self
        import hotssh.version
        dialog.set_property('website', 'http://www.gnome.org/projects/hotssh')
        dialog.set_property('version', hotssh.version.__version__)
        dialog.set_property('authors', ['Colin Walters <walters@verbum.org>'])
        dialog.set_property('copyright', u'Copyright \u00A9 2007,2008 Colin Walters <walters@verbum.org>')
        dialog.set_property('logo-icon-name', 'hotwire-openssh')
        dialog.set_property('license',
                            '''Hotwire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.\n
Hotwire is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.\n
You should have received a copy of the GNU General Public License
along with Hotwire; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA''')
        dialog.set_property('name', "About Secure Shell")
        comments = _("An interface to OpenSSH, a Secure Shell client\n\n")
        if hotssh.version.svn_version_info:
            comments += "changeset: %s\ndate: %s\n" % (hotssh.version.svn_version_info['Revision'], hotssh.version.svn_version_info['Last Changed Date'],)
        dialog.set_property('comments', comments)

class SshConnectionHistory(object):
    def __init__(self):
        # We want to upgrade from the old Hotwire-derived location into the new ~/.hotssh
        self.__oldstatedir = os.path.expanduser('~/.hotwire/state')
        self.__statedir = os.path.expanduser('~/.hotssh')
        try:
            os.makedirs(self.__statedir)
        except:
            pass
        self.__path = path = os.path.join(self.__statedir, 'history.sqlite')
        oldpath = os.path.join(self.__oldstatedir, 'ssh.sqlite')
        if os.path.exists(oldpath) and not os.path.exists(self.__path):
            shutil.copy2(oldpath, path)
        self.__conn = sqlite3.connect(path, isolation_level=None)
        cursor = self.__conn.cursor()
        cursor.execute('''CREATE TABLE IF NOT EXISTS Connections (bid INTEGER PRIMARY KEY AUTOINCREMENT, host TEXT, user TEXT, options TEXT, conntime DATETIME)''')
        cursor.execute('''CREATE INDEX IF NOT EXISTS ConnectionsIndex1 on Connections (host)''')
        cursor.execute('''CREATE INDEX IF NOT EXISTS ConnectionsIndex2 on Connections (host,user)''')
        cursor.execute('''CREATE TABLE IF NOT EXISTS Colors (bid INTEGER PRIMARY KEY AUTOINCREMENT, pattern TEXT UNIQUE, fg TEXT, bg TEXT, modtime DATETIME)''')
        cursor.execute('''CREATE INDEX IF NOT EXISTS ColorsIndex on Colors (pattern)''')

    def get_recent_connections_search(self, host, limit=10):
        cursor = self.__conn.cursor()
        params = ()
        q = '''SELECT user,host,options,conntime FROM Connections '''
        if host:
            q += '''WHERE host LIKE ? ESCAPE '%' '''
            params = ('%' + host.replace('%', '%%') + '%',)
        # We use a large limit here to avoid the query being totally unbounded
        q += '''ORDER BY conntime DESC LIMIT 1000'''
        def sqlite_timestamp_to_datetime(s):
            return datetime.datetime.fromtimestamp(time.mktime(time.strptime(s[:-7], '%Y-%m-%d %H:%M:%S')))
        # Uniquify, which we coudln't do in the SQL because we're also grabbing the conntime
        seen = set()
        for user,host,options,conntime in cursor.execute(q, params):
            if len(seen) >= limit:
                break
            if user and host:
                uhost = user+'@'+host
            else:
                uhost = host
            if uhost in seen:
                continue
            seen.add(uhost)
            # We do this just to parse conntime
            yield user,host,options,sqlite_timestamp_to_datetime(conntime)

    def get_users_for_host_search(self, host):
        for user,host,options,ts in self.get_recent_connections_search(host):
            yield user

    def add_connection(self, user, host, options):
        if host is None or host == '':
            return
        cursor = self.__conn.cursor()
        cursor.execute('''BEGIN TRANSACTION''')
        cursor.execute('''INSERT INTO Connections VALUES (NULL, ?, ?, ?, ?)''', (host, user, ' '.join(options), datetime.datetime.now()))
        cursor.execute('''COMMIT''')

    def get_color_for_host(self, host):
        cursor = self.__conn.cursor()
        res = cursor.execute('SELECT fg,bg FROM Colors WHERE pattern = ? ORDER BY conntime DESC LIMIT 10')
        if len(res) > 0:
            return res[0]
        return None

    def set_color_for_host(self, host, fg, bg):
        cursor.execute('''BEGIN TRANSACTION''')
        cursor.execute('''INSERT INTO Colors VALUES (NULL, ?, ?, ?, ?)''', (host, fg, bg, datetime.datetime.now()))
        cursor.execute('''COMMIT''')
_history_instance = None
def get_history():
    global _history_instance
    if _history_instance is None:
        _history_instance = SshConnectionHistory()
    return _history_instance

class OpenSSHKnownHostsDB(gobject.GObject):
    __gsignals__ = {
        "changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
    }
    def __init__(self):
        super(OpenSSHKnownHostsDB, self).__init__()
        self.__path = os.path.expanduser('~/.ssh/known_hosts')
        self.__hostcache_ts_size = (None, None)
        self.__hostcache = None
        self.__faviconcache = os.path.expanduser('~/.hotssh/favicons')
        self.__pixbufcache = {}

    def __get_favicon_cache(self):
        try:
            os.makedirs(self.__faviconcache)
        except OSError, e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise
        return self.__faviconcache

    def __read_hosts(self):
        try:
            _logger.debug("reading %s", self.__path)
            f = open(self.__path)
        except:
            _logger.debug("failed to open known hosts")
            return
        hosts = {}
        for line in f:
            if not line or line.startswith('|'):
                continue
            hostip,rest = line.split(' ', 1)
            if hostip.find(',') > 0:
                host = hostip.split(',', 1)[0]
            else:
                host = hostip
            hostkey_type,hostkey = rest.split(' ', 1)
            hostkey_type = hostkey_type.strip()
            hostkey = hostkey.strip()
            host = host.strip()
            hosts[host] = (hostkey_type, hostkey)
        f.close()
        return hosts

    def get_hosts(self):
        try:
            stbuf = os.stat(self.__path)
            ts_size =  (stbuf[stat.ST_MTIME], stbuf[stat.ST_SIZE])
        except OSError, e:
            ts_size = None
        if ts_size is not None and self.__hostcache_ts_size != ts_size:
            self.__hostcache = self.__read_hosts()
            self.__hostcache_ts_size = ts_size
        else:
            return []
        return self.__hostcache.iterkeys()

    def __force_host_read(self):
        self.get_hosts()

    def get_favicon_for_host(self, host, port=None):
        cache = self.__get_favicon_cache()
        if cache is None:
            return None
        path = self.__get_favicon_path(host, port)
        try:
            stbuf = os.stat(path)
        except:
            return None
        return (path, stbuf[stat.ST_MTIME])

    def render_cached_favicon(self, path):
        bn = os.path.basename(path)
        if bn not in self.__pixbufcache:
            self.__pixbufcache[bn] = gtk.gdk.pixbuf_new_from_file_at_size(path, 16, 16)
        return self.__pixbufcache[bn]

    def __get_favicon_path(self, hostname, port):
        key = urllib.quote_plus(hostname)
        return os.path.join(self.__get_favicon_cache(), key + '.png')

    def save_favicon(self, host, port, favicon_tmppath):
        cache = self.__get_favicon_cache()
        if cache is None:
            return None
        path = self.__get_favicon_path(host, port)
        shutil.move(favicon_tmppath, path)
        self.emit('changed')
        return path

_openssh_hosts_db = OpenSSHKnownHostsDB()

class SshAvahiMonitor(gobject.GObject):
    __gsignals__ = {
        "changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
    }
    def __init__(self):
        super(SshAvahiMonitor, self).__init__()
        # maps name->(addr,port)
        self.__cache = {}

        if not avahi_available:
            return

        try:
            sysbus = dbus.SystemBus()
            avahi_service = sysbus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER)
            self.__avahi = dbus.Interface(avahi_service, avahi.DBUS_INTERFACE_SERVER)
        except dbus.exceptions.DBusException:
            return

        browser_ref = self.__avahi.ServiceBrowserNew(avahi.IF_UNSPEC,
                                                     avahi.PROTO_UNSPEC,
                                                     '_ssh._tcp', 'local',
                                                     dbus.UInt32(0))
        browser = sysbus.get_object(avahi.DBUS_NAME, browser_ref)
        self.__browser = dbus.Interface(browser, avahi.DBUS_INTERFACE_SERVICE_BROWSER)
        self.__browser.connect_to_signal('ItemNew', self.__on_item_new)
        self.__browser.connect_to_signal('ItemRemove', self.__on_item_remove)

    def __iter__(self):
        for k,(addr,port) in self.__cache.iteritems():
            yield (k, addr, port)

    def __on_resolve_reply(self, *args):
        addr, port = args[-4:-2]
        name = args[2].decode('utf-8', 'replace')
        _logger.debug("add Avahi SSH: %r %r %r", name, addr, port)
        self.__cache[name] = (addr, port)
        self.emit('changed')

    def __on_avahi_error(self, *args):
        _logger.debug("Avahi error: %s", args)

    def __on_item_new(self, interface, protocol, name, type, domain, flags):
        self.__avahi.ResolveService(interface, protocol, name, type, domain,
                                    avahi.PROTO_UNSPEC, dbus.UInt32(0),
                                    reply_handler=self.__on_resolve_reply,
                                    error_handler=self.__on_avahi_error)

    def __on_item_remove(self, interface, protocol, name, type, domain,server):
        uname = name.decode('utf-8', 'replace')
        if uname in self.__cache:
            _logger.debug("del Avahi SSH: %r", uname)
            del self.__cache[uname]
            self.emit('changed')

_local_avahi_instance = None
def get_local_avahi():
    global _local_avahi_instance
    if _local_avahi_instance is None:
        _local_avahi_instance = SshAvahiMonitor()
    return _local_avahi_instance

class SshOptions(gtk.Expander):
    def __init__(self):
        super(SshOptions, self).__init__(_('Options'))
        self.set_label('<b>%s</b>' % (gobject.markup_escape_text(self.get_label())))
        self.set_use_markup(True)
        options_vbox = gtk.VBox()
        options_hbox = gtk.HBox()
        options_vbox.add(options_hbox)
        self.__entry = gtk.Entry()
        options_hbox.pack_start(self.__entry, expand=True)
        options_help = gtk.Button(stock=gtk.STOCK_HELP)
        options_help.connect('clicked', self.__on_options_help_clicked)
        options_hbox.pack_start(options_help, expand=False)
        options_helplabel = gtk.Label(_('Example: '))
        options_helplabel.set_markup('<i>%s -C -Y -oTCPKeepAlive=false</i>' % (gobject.markup_escape_text(options_helplabel.get_label()),))
        options_helplabel.set_alignment(0.0, 0.5)
        options_vbox.add(options_helplabel)
        self.add(options_vbox)

    def get_entry(self):
        return self.__entry

    def __on_options_help_clicked(self, b):
        # Hooray Unix!
        subprocess.Popen(['gnome-terminal', '-x', 'man', 'ssh'])

class ConnectDialog(gtk.Dialog):
    def __init__(self, parent=None, history=None, local_avahi=None):
        super(ConnectDialog, self).__init__(title=_("New Secure Shell Connection"),
                                            parent=parent,
                                            flags=gtk.DIALOG_DESTROY_WITH_PARENT,
                                            buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        self.__history = history
        self.__local_avahi = local_avahi
        self.__local_avahi.connect('changed', self.__on_local_avahi_changed)

        self.__default_username = pwd.getpwuid(os.getuid()).pw_name

        self.connect('response', lambda *args: self.hide())
        self.connect('delete-event', self.hide_on_delete)
        button = self.add_button(_('C_onnect'), gtk.RESPONSE_ACCEPT)
        button.set_property('image', gtk.image_new_from_stock('gtk-connect', gtk.ICON_SIZE_BUTTON))
        self.set_default_response(gtk.RESPONSE_ACCEPT)

        self.set_has_separator(False)
        self.set_border_width(5)

        self.__vbox = vbox = gtk.VBox()
        self.vbox.add(self.__vbox)
        self.vbox.set_spacing(6)

        self.__response_value = None

        self.__viewmode = 'history'

        self.__suppress_recent_search = False
        self.__idle_search_id = 0
        self.__idle_update_search_id = 0

        self.__custom_user = False

        self.__hosts = _openssh_hosts_db

        sg = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)

        hbox = gtk.HBox()
        vbox.pack_start(hbox, expand=False)
        host_label = gtk.Label(_('Host: '))
        host_label.set_markup('<b>%s</b>' % (gobject.markup_escape_text(host_label.get_text())))
        sg.add_widget(host_label)
        hbox.pack_start(host_label, expand=False)
        self.__entry = gtk.combo_box_entry_new_text()
        self.__entry.child.connect('activate', self.__on_entry_activated)
        self.__entry.child.connect('notify::text', self.__on_entry_modified)
        self.__entrycompletion = gtk.EntryCompletion()
        self.__entrycompletion.set_property('inline-completion', True)
        self.__entrycompletion.set_property('popup-completion', False)
        self.__entrycompletion.set_property('popup-single-match', False)
        self.__entrycompletion.set_model(self.__entry.get_property('model'))
        self.__entrycompletion.set_text_column(0)
        self.__entry.child.set_completion(self.__entrycompletion)
        hbox.add(self.__entry)
        self.__reload_entry()

        hbox = gtk.HBox()
        vbox.pack_start(hbox, expand=False)
        user_label = gtk.Label(_('User: '))
        user_label.set_markup('<b>%s</b>' % (gobject.markup_escape_text(user_label.get_text())))
        sg.add_widget(user_label)
        hbox.pack_start(user_label, expand=False)
        self.__user_entry = gtk.Entry()
        self.__set_user(None)
        self.__user_entry.connect('notify::text', self.__on_user_modified)
        self.__user_entry.set_activates_default(True)
        self.__user_completion = gtk.EntryCompletion()
        self.__user_completion.set_property('inline-completion', True)
        self.__user_completion.set_property('popup-single-match', False)
        self.__user_completion.set_model(gtk.ListStore(gobject.TYPE_STRING))
        self.__user_completion.set_text_column(0)
        self.__user_entry.set_completion(self.__user_completion)

        hbox.pack_start(self.__user_entry, expand=False)

        vbox.pack_start(gtk.Label(' '), expand=False)

        self.__conntype_notebook = gtk.Notebook()
        self.__conntype_notebook.set_scrollable(True)
        self.__conntype_notebook.connect('switch-page', self.__on_switch_page)
        vbox.pack_start(self.__conntype_notebook, expand=True)

        history_label = gtk.Label(_('History'))
        tab = gtk.ScrolledWindow()
        tab.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        #frame.set_label_widget(history_label)
        self.__recent_model = gtk.ListStore(object, str,  object)
        self.__recent_view = gtk.TreeView(self.__recent_model)
        self.__recent_view.get_selection().connect('changed', self.__on_recent_selected)
        self.__recent_view.connect('row-activated', self.__on_recent_activated)
        tab.add(self.__recent_view)
        colidx = self.__recent_view.insert_column_with_data_func(-1, '',
                                                          gtk.CellRendererPixbuf(),
                                                          self.__render_favicon)
        colidx = self.__recent_view.insert_column_with_data_func(-1, _('Connection'),
                                                          gtk.CellRendererText(),
                                                          self.__render_userhost)
        colidx = self.__recent_view.insert_column_with_data_func(-1, _('Time'),
                                                          gtk.CellRendererText(),
                                                          self.__render_time_recency, datetime.datetime.now())

        self.__conntype_notebook.append_page(tab, history_label)

        local_label = gtk.Label(_('Local'))
        tab = gtk.ScrolledWindow()
        tab.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        self.__local_model = gtk.ListStore(str, str, str)
        self.__local_view = gtk.TreeView(self.__local_model)
        self.__local_view.get_selection().connect('changed', self.__on_local_selected)
        self.__local_view.connect('row-activated', self.__on_local_activated)
        tab.add(self.__local_view)
        colidx = self.__local_view.insert_column_with_attributes(-1, _('Name'),
                                                          gtk.CellRendererText(),
                                                          text=0)
        colidx = self.__local_view.insert_column_with_attributes(-1, _('Address'),
                                                          gtk.CellRendererText(),
                                                          text=1)
        self.__conntype_notebook.append_page(tab, local_label)

        self.__on_entry_modified()
        self.__force_idle_search()

        self.__options_expander = SshOptions()
        self.__options_entry = self.__options_expander.get_entry()

        vbox.pack_start(self.__options_expander, expand=False)

        self.set_default_size(640, 480)

    @log_except(_logger)
    def __on_switch_page(self, nb, p, pn):
        _logger.debug("got page switch, pn=%d", pn)
        self.__suppress_recent_search = True
        widget = self.__conntype_notebook.get_nth_page(pn)
        self.__viewmode = (pn == 0 and 'history' or 'local')
        # If we're switching, particularly from history->local it's
        # unlikely we want to keep the active search; the assumption here
        # is that the history search failed.
        self.__entry.child.set_text('')
        self.__force_idle_search()
        self.__suppress_recent_search = False

    def __on_browse_local(self, b):
        pass

    def __on_local_avahi_changed(self, *args):
        if self.__viewmode == 'local':
            self.__force_idle_search()

    def __render_userhost(self, col, cell, model, iter):
        user = model.get_value(iter, 0)
        host = model.get_value(iter, 1)
        if user and host:
            userhost = user + '@' + host
        else:
            userhost = host
        cell.set_property('text', userhost)

    def __set_user(self, name):
        if name is None:
            name = self.__default_username
        self.__user_entry.set_text(name)

    def __render_favicon(self, col, cell, model, it):
        user = model.get_value(it, 0)
        host = model.get_value(it, 1)
        favicondata = _openssh_hosts_db.get_favicon_for_host(host, None)
        if favicondata is not None:
            (favicon_path,mtime) = favicondata
            pixbuf = _openssh_hosts_db.render_cached_favicon(favicon_path)
            cell.set_property('pixbuf',pixbuf)
        else:
            cell.set_property('pixbuf', None)

    def __render_time_recency(self, col, cell, model, it, curtime):
        val = model.get_value(it, 2)
        deltastr = timesince(val, curtime)
        cell.set_property('text', deltastr)

    def __reload_entry(self, *args, **kwargs):
        _logger.debug("reloading")
        # TODO do predictive completion here
        # For example, I have in my history:
        # foo.cis.ohio-state.edu
        # bar.cis.ohio-state.edu
        # Now I type baz.cis
        # The system notices that nothing matches baz.cis; however
        # the "cis" part does match foo.cis.ohio-state.edu, so
        # we start offering a completion for it
        self.__entry.get_property('model').clear()
        for host in self.__hosts.get_hosts():
            self.__entry.append_text(host)

    def __on_user_modified(self, *args):
        text = self.__user_entry.get_text()
        self.__custom_user = text != self.__default_username

    def __on_entry_modified(self, *args):
        text = self.__entry.get_active_text()
        havetext = text != ''
        self.set_response_sensitive(gtk.RESPONSE_ACCEPT, havetext)
        if self.__suppress_recent_search:
            return
        if havetext and self.__idle_update_search_id == 0:
            self.__idle_update_search_id = gobject.timeout_add(250, self.__idle_update_search)

    def __force_idle_search(self):
        if self.__idle_update_search_id > 0:
            gobject.source_remove(self.__idle_update_search_id)
        self.__idle_update_search()

    def __idle_update_search(self):
        self.__idle_update_search_id = 0
        self.__suppress_recent_search = True
        if self.__viewmode == 'history':
            self.__idle_update_search_history()
        else:
            self.__idle_update_search_local()
        self.__suppress_recent_search = False

    def __idle_update_search_local(self):
        self.__local_view.get_selection().unselect_all()
        hosttext = self.__entry.get_active_text()
        self.__local_model.clear()
        for name,host,port in self.__local_avahi:
            if host.find(hosttext) >= 0:
                self.__local_model.append((name,host,port))

    def __idle_update_search_history(self):
        self.__recent_view.get_selection().unselect_all()
        host = self.__entry.get_active_text()
        if host:
            usernames = list(self.__history.get_users_for_host_search(host))
        else:
            usernames = []
        _logger.debug("doing history update, usernames:%r", usernames)
        if len(usernames) > 0:
            last_user = usernames[0]
            if last_user is None:
                last_user = self.__default_username
            self.__user_entry.set_text(last_user)
            model = gtk.ListStore(gobject.TYPE_STRING)
            for uname in usernames:
                if uname:
                    model.append((uname,))
            self.__user_completion.set_model(model)
        else:
            self.__user_entry.set_text(self.__default_username)
        self.__reload_connection_history()

    def __reload_connection_history(self):
        hosttext = self.__entry.get_active_text()
        self.__recent_model.clear()
        for user,host,options,ts in self.__history.get_recent_connections_search(hosttext):
            self.__recent_model.append((user, host, ts))

    def __on_entry_activated(self, *args):
        self.__force_idle_search()
        hosttext = self.__entry.get_active_text()
        if hosttext.find('@') >= 0:
            (user, host) = hosttext.split('@', 1)
            self.__user_entry.set_text(user)
            self.__entry.child.set_text(host)
            self.__user_entry.activate()
        else:
            self.__user_entry.select_region(0, -1)
            self.__user_entry.grab_focus()

    def __on_local_selected(self, ts):
        if self.__suppress_recent_search:
            return
        _logger.debug("local selected: %s", ts)
        (tm, seliter) = ts.get_selected()
        if seliter is None:
            return
        host = self.__local_model.get_value(seliter, 1)
        port = self.__local_model.get_value(seliter, 2)
        self.__suppress_recent_search = True
        self.__entry.child.set_text(host)
        self.__suppress_recent_search = False

    def __on_local_activated(self, tv, path, vc):
        self.activate_default()

    def __on_recent_selected(self, ts):
        if self.__suppress_recent_search:
            return
        _logger.debug("recent selected: %s", ts)
        (tm, seliter) = ts.get_selected()
        if seliter is None:
            return
        user = self.__recent_model.get_value(seliter, 0)
        host = self.__recent_model.get_value(seliter, 1)
        self.__suppress_recent_search = True
        self.__entry.child.set_text(host)
        if user:
            self.__user_entry.set_text(user)
        else:
            self.__set_user(None)
        self.__suppress_recent_search = False

    def __on_recent_activated(self, tv, path, vc):
        self.activate_default()

    def __reset(self):
        self.__entry.child.set_text('')
        self.__user_entry.set_text(self.__default_username)

    def run_get_cmd(self):
        self.__reset()
        self.show_all()
        resp = self.run()
        if resp != gtk.RESPONSE_ACCEPT:
            return None
        host = self.__entry.get_active_text()
        if not host:
            return None
        args = [x for x in _whitespace_re.split(self.__options_entry.get_text()) if x != '']
        if self.__custom_user:
            args.append(self.__user_entry.get_text() + '@' + host)
        else:
            args.append(host)
        return args

_CONTROLPATH = None
def get_controlpath(create=True):
    global _CONTROLPATH
    if _CONTROLPATH is None and create:
        _CONTROLPATH = tempfile.mkdtemp('', 'hotssh')
    return _CONTROLPATH

def get_base_sshcmd():
    return ['ssh']

def get_connection_sharing_args():
    # TODO - openssh should really do this out of the box
    return ['-oControlMaster=auto', '-oControlPath=' + os.path.join(get_controlpath(), 'master-%r@%h:%p')]

class AsyncCommandWithOutput(gobject.GObject):
    __gsignals__ = {
        "timeout" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_STRING]),

        "complete" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (bool,
                                                                   int,
                                                                   gobject.TYPE_STRING))
    }

    user = property(lambda self: self.__user)
    host = property(lambda self: self.__host)
    port = property(lambda self: self.__port)
    output = property(lambda self: self.__output.get_value())

    def __init__(self, user, host, port, cmdstart, cmdend, timeout=7000, hostappend='', getoutput=True):
        super(AsyncCommandWithOutput, self).__init__()
        self.__user = user
        self.__host = host
        self.__port = port
        cmd = list(cmdstart)
        cmd.extend(get_connection_sharing_args())
        if port:
            cmd.extend(['-p', str(port)])
        if user:
            cmd.append('-oUser=' + user)
        # hostappend is a dirty hack to handle scp's syntax
        cmd.extend(['-oBatchMode=true', host+hostappend])
        cmd.extend(cmdend)
        self.__cmd = cmd
        self.__starttime = time.time()
        nullf = open(os.path.devnull, 'w')
        _logger.debug("starting subprocess cmd=%r", cmd)
        if getoutput:
            stdout_target = subprocess.PIPE
        else:
            stdout_target = nullf
        self.__subproc = subprocess.Popen(cmd, stdout=stdout_target, stderr=nullf)
        nullf.close()
        if getoutput:
            self.__io_watch_id = gobject.io_add_watch(self.__subproc.stdout,
                                                      gobject.IO_IN|gobject.IO_ERR|gobject.IO_HUP,
                                                      self.__on_io)
        self.__output = StringIO()
        self.__child_watch_id = gobject.child_watch_add(self.__subproc.pid, self.__on_exited)
        self.__timeout_id = gobject.timeout_add(timeout, self.__on_timeout)

    def __on_io(self, source, condition):
        have_read = condition & gobject.IO_IN
        if have_read:
            _logger.debug("got status output")
            self.__output.write(os.read(source.fileno(), 8192))
        if ((condition & gobject.IO_HUP) or (condition & gobject.IO_ERR)):
            source.close()
            _logger.debug("got condition %s, cancelling status io check", condition)
            return False
        else:
            return have_read

    def __on_timeout(self):
        _logger.debug("timeout for host=%r cmd=%r", self.__host, self.__cmd)
        try:
            os.kill(self.__subproc.pid, signal.SIGHUP)
        except OSError, e:
            _logger.debug("failed to signal pid %s", pid, exc_info=True)
            pass
        self.emit('timeout', self.__output.getvalue())
        self.__timeout_id = 0
        return False

    def __on_exited(self, pid, condition):
        _logger.debug("command exited host=%r condition=%r cmd=%r", self.__host, condition, self.__cmd)
        if self.__timeout_id == 0:
            _logger.debug("command exited (but timeout already run")
            return
        gobject.source_remove(self.__timeout_id)
        self.__timeout_id = 0
        self.emit('complete', condition == 0, time.time() - self.__starttime, self.__output.getvalue())
        return False

class FaviconRetriever(gobject.GObject):
    __gsignals__ = {
        "favicon-loaded" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [str]),
    }

    user = property(lambda self: self.__user)
    host = property(lambda self: self.__host)
    port = property(lambda self: self.__port)

    def __init__(self, user, host, port=None):
        super(FaviconRetriever, self).__init__()
        self.__user = user
        self.__host = host
        self.__port = port
        self.__cached_mtime = None
        self.__favicon_mtime = None
        self.__tmp_favicon_path = None

        hostport = userhost_pair_to_string(user, host, port)
        _logger.debug("loading favicon for %s", hostport)
        cmd = commands.mkarg('''python -c 'import os,sys,stat; print os.stat(sys.argv[1])[stat.ST_MTIME]' /etc/favicon.png''')
        req = AsyncCommandWithOutput(user, host, port, ['ssh'], ['sh', '-c', cmd])
        req.connect('timeout', self.__on_mtime_timeout)
        req.connect('complete', self.__on_mtime_complete)
        self.__active_req = req

    def __on_mtime_timeout(self, req, curoutput):
        _logger.debug("favicon mtime retrieval timeout")
        return False

    def __on_mtime_complete(self, req, status, elapsed_time, output):
        _logger.debug("favicon mtime retrieval complete, status: %r", status)
        if not status:
            return False
        mtime = int(output.strip())
        self.__favicon_mtime = mtime
        gobject.idle_add(self.__idle_start_scp)

    @log_except(_logger)
    def __idle_start_scp(self):
        current = _openssh_hosts_db.get_favicon_for_host(self.__host, self.__port)
        cached_mtime = 0
        if current is not None:
            (_, mtime) = current
            cached_mtime = mtime
        if self.__favicon_mtime > cached_mtime:
            (fd, tmppath) = tempfile.mkstemp('.png', 'favicon')
            self.__tmp_favicon_path = tmppath
            os.close(fd)

            _logger.debug("creating scp request")
            req = AsyncCommandWithOutput(self.__user, self.__host, self.__port,
                                         ['scp', '-p', '-q'], [tmppath],
                                         hostappend=':/etc/favicon.png', timeout=15000, getoutput=False)
            req.connect('timeout', self.__on_favicon_timeout)
            req.connect('complete', self.__on_favicon_complete)
            self.__active_req = req
        else:
            _logger.debug("favicon is up to date")
        return False

    def __on_favicon_timeout(self, req, elapsed):
        _logger.debug("favicon data retrieval timeout (%r elapsed)", elapsed)
        return False

    def __on_favicon_complete(self, req, status, elapsed_time, output):
        _logger.debug("favicon data retrieval complete; elapsed=%r status=%r", elapsed_time, status)
        if not status:
            return False
        self.emit('favicon-loaded', self.__tmp_favicon_path)
        return False

class HostConnectionMonitor(gobject.GObject):
    __gsignals__ = {
        "host-status" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,
                                                                      gobject.TYPE_BOOLEAN,
                                                                      gobject.TYPE_PYOBJECT,
                                                                      gobject.TYPE_PYOBJECT)),
    }
    def __init__(self):
        super(HostConnectionMonitor, self).__init__()
        
        self._CHECK_MS = 15000
        
        # map userhost -> failure count
        self._host_failures = {}
        # map userhost -> ids
        self.__host_monitor_ids = {}
        # map userhost -> various idle ids, etc
        self.__check_statuses = {}
        # Map userhost -> w output
        self.__check_status_output = {}
        

    def start_monitor(self, user, host):
        userhost = userhost_pair_to_string(user, host)
        if not (userhost in self.__host_monitor_ids or userhost in self.__check_statuses):
            _logger.debug("adding monitor for %s", userhost)
            self._host_failures[userhost] = 0
            self.__host_monitor_ids[userhost] = gobject.timeout_add(self._CHECK_MS, self.__check_host, userhost)

    def stop_monitor(self, user, host):
        userhost = userhost_pair_to_string(user, host)        
        _logger.debug("stopping monitor for %s", userhost)
        if userhost in self.__host_monitor_ids:
            monid = self.__host_monitor_ids[userhost]
            gobject.source_remove(monid)
            del self.__host_monitor_ids[userhost]
        if userhost in self.__check_statuses:
            del self.__check_statusesuser[userhost]

    def get_monitors(self):
        return self.__host_monitor_ids

    def __on_check_io(self, source, condition, userhost):
        have_read = condition & gobject.IO_IN
        if have_read:
            _logger.debug("got status output")
            self.__check_status_output[userhost] += os.read(source.fileno(), 8192)
        if ((condition & gobject.IO_HUP) or (condition & gobject.IO_ERR)):
            source.close()
            _logger.debug("got condition %s, cancelling status io check", condition)
            return False
        else:
            return have_read

    def __check_host(self, userhost):
        _logger.debug("performing check for %s", userhost)
        del self.__host_monitor_ids[userhost]
        cmd = list(get_base_sshcmd())
        cmd.extend(get_connection_sharing_args())
        starttime = time.time()
        # This is a hack.  Blame Adam Jackson.
        cmd.extend(['-oBatchMode=true', userhost, 'uptime'])
        nullf = open(os.path.devnull, 'w')
        subproc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=nullf)
        nullf.close()
        io_watch_id = gobject.io_add_watch(subproc.stdout,
                                           gobject.IO_IN|gobject.IO_ERR|gobject.IO_HUP,
                                           self.__on_check_io, userhost)
        self.__check_status_output[userhost] = ""
        child_watch_id = gobject.child_watch_add(subproc.pid, self.__on_check_exited, userhost)
        timeout_id = gobject.timeout_add(7000, self.__check_timeout, userhost)
        self.__check_statuses[userhost] = [starttime, subproc.pid, timeout_id, child_watch_id, io_watch_id]
        return False

    def __check_timeout(self, userhost):
        _logger.debug("timeout for host=%r", userhost)
        try:
            (starttime, pid, timeout_id, child_watch_id, io_watch_id) = self.__check_statuses[userhost]
        except KeyError, e:
            return False
        self.__check_statuses[2] = 0
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError, e:
            _logger.debug("failed to signal pid %s", pid, exc_info=True)
            pass
        return False

    def __on_check_exited(self, pid, condition, userhost):
        _logger.debug("check exited, pid=%s condition=%s host=%s", pid, condition, userhost)
        try:
            (starttime, pid, timeout_id, child_watch_id, io_watch_id) = self.__check_statuses[userhost]
        except KeyError, e:
            _logger.debug("confused; no status for userhost=%s", userhost)
            return False
        if timeout_id > 0:
            gobject.source_remove(timeout_id)
        del self.__check_statuses[userhost]
        abort = False
        if condition == 0:
            self._host_failures[userhost] = 0
        else:
            self._host_failures[userhost] += 1
            if self._host_failures[userhost] > 3:
                _logger.debug("too many failures for host=%s", userhost)
                abort = True
        if not abort:
            self.__host_monitor_ids[userhost] = gobject.timeout_add(self._CHECK_MS, self.__check_host, userhost)
        self.emit('host-status', userhost, condition == 0, time.time()-starttime, self.__check_status_output[userhost])
        del self.__check_status_output[userhost]
        return False

_hostmonitor = HostConnectionMonitor()

class SshTerminalWidgetImpl(VteTerminalWidget):
    def __init__(self, *args, **kwargs):
        self.__actions = kwargs['actions']
        super(SshTerminalWidgetImpl, self).__init__(*args, **kwargs)

    def _get_extra_context_menuitems(self):
        return [self.__actions.get_action(x).create_menu_item() for x in ['CopyConnection', 'OpenSFTP']]

class SshTerminalWidget(gtk.VBox):
    __gsignals__ = {
        "status-changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
        "close" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
        # Emitted when we do a reconnect so that other tabs for this host can pick it up
        "reconnect" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
        "metadata-changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
    }

    latency = property(lambda self: self.__latency)
    status_line = property(lambda self: self.__status_output_first)
    connecting_state = property(lambda self: self.__connecting_state)
    connected = property(lambda self: self.__connected)
    ssh_options = property(lambda self: self.__sshopts)

    ssh_opt_with_args = "bcDeFiLlmOopRSw"
    ssh_opt_witouth_args = "1246AaCfgKkMNnqsTtVvXxY"

    def __init__(self, args, cwd, actions=None, inituser=None, inithost=None):
        super(SshTerminalWidget, self).__init__()

        self.__args = args
        self.__sshcmd = list(get_base_sshcmd())
        self.__cwd = cwd
        self.__user = inituser
        self.__host = inithost
	self.__port = None
        self.__sshopts = []
        self.__actions = actions
        self.__idle_start_favicon_id = 0
        self.__favicon_path = None
        self.__favicon_pixbuf = None
        self.__favicon_retriever = None
        self.__favicon_retriever_connections = []

        self.__init_state()

        enable_connection_sharing = True
        need_arg = False
        this_is_port = False
        port_in_args = False

        for arg in args:
            if not need_arg and not arg.startswith('-'):
                need_arg = False
                if self.__host is None:
                    host_port = arg.split(":")
                    if len(host_port) == 2:
                        self.__host = host_port[0]
                        self.__port = host_port[1]
                    else:
                        self.__host = arg
                    try:
                        (self.__user,self.__host) = self.__host.split('@', 1)
                    except ValueError, e:
                        pass
            elif this_is_port:
                self.__port = arg
                this_is_port = False
                need_arg = False
                port_in_args = True
                self.__sshcmd.append(arg)
            else:
                if arg == '-oControlPath=none':
                    enable_connection_sharing = False
                elif self.ssh_opt_with_args.find(arg[1:]) != -1:
                    need_arg = True
                    if arg == "-p":
                        this_is_port = True
                else:
                    need_arg = False
                self.__sshcmd.append(arg)

        if self.__user:
            self.__sshcmd.append('-oUser=' + self.__user)
        self.__sshcmd.append(self.__host)

        if not port_in_args and self.__port:
            self.__sshcmd.append("-p")
            self.__sshcmd.append(self.__port)

        if enable_connection_sharing:
            self.__sshcmd.extend(get_connection_sharing_args())

        header = gtk.VBox()
        self.__msg = gtk.Label()
        self.__msg.set_alignment(0.0, 0.5)
        self.__msgarea_mgr = MsgAreaController()
        #header.pack_start(self.__msg)
        header.pack_start(self.__msgarea_mgr)
        self.pack_start(header, expand=False)
        self.ssh_connect()

    def __init_state(self):
        if self.__idle_start_favicon_id > 0:
            gobject.source_remove(self.__idle_start_favicon_id)
            self.__idle_start_favicon_id = 0
        if self.__favicon_retriever is not None:
            for sigid in self.__favicon_retriever_connections:
                self.__favicon_retriever.disconnect(sigid)
            self.__favicon_retriever_connections = []
        self.__favicon_retriever = None

        self.__global_connection_changed = False
        self.__connecting_state = False
        self.__connected = None
        self.__cmd_exited = False
        self.__latency = None
        self.__status_output_first = None

    def set_global_connection_changed(self, changed):
        self.__global_connection_changed = changed
        if changed:
            self.__msgarea_mgr.clear()
        elif self.__cmd_exited:
            self.__show_disconnected()

    def set_status(self, connected, latency, status_output=None):
        if not connected and self.__connecting_state:
            return
        self.__connecting_state = False
        connected_changed = self.__connected != connected
        latency_changed = (not self.__latency) or (self.__latency*0.9 > latency) or (self.__latency*1.1 < latency)
        if not (connected_changed or latency_changed):
            return
        self.__connected = connected
        self.__latency = latency
        buf = StringIO(status_output)
        firstline = buf.readline().strip()
        self.__status_output_first = firstline
        self.emit('status-changed')
        self.__sync_msg()

    def __sync_msg(self):
        return

    def __on_favicon_loaded(self, retriever, faviconpath):
        _logger.debug("favicon loaded; path=%r", faviconpath)
        self.__favicon_path = _openssh_hosts_db.save_favicon(self.__host, self.__port, faviconpath)
        self.__favicon_pixbuf = _openssh_hosts_db.render_cached_favicon(self.__favicon_path)
        self.emit('metadata-changed')

    def ssh_connect(self):
        self.__connecting_state = True
        self.__term = term = SshTerminalWidgetImpl(cwd=self.__cwd, cmd=self.__sshcmd, actions=self.__actions)
        term.connect('child-exited', self.__on_child_exited)
        term.show_all()
        self.pack_start(term, expand=True)
        # For some reason, VTE doesn't have the CAN_FOCUS flag set, so we can't call
        # grab_focus.  Do it manually then:
        term.emit('focus', True)
        self.__msgarea_mgr.clear()
        self.__sync_msg()

        current = _openssh_hosts_db.get_favicon_for_host(self.__host, self.__port)
        if current is not None:
            (pixbufpath, mtime) = current
            self.__favicon_path = pixbufpath
            self.__favicon_pixbuf = _openssh_hosts_db.render_cached_favicon(pixbufpath)
        self.__idle_start_favicon_id = gobject.timeout_add(3000, self.__idle_start_favicon)

    def __idle_start_favicon(self):
        if self.__favicon_retriever is not None:
            return
        self.__favicon_retriever = FaviconRetriever(self.__user, self.__host, self.__port)
        sigid = self.__favicon_retriever.connect('favicon-loaded', self.__on_favicon_loaded)
        self.__favicon_retriever_connections.append(sigid)

    def ssh_reconnect(self):
        # TODO - do this in a better way
        if not self.__term.exited and self.__term.pid:
            os.kill(self.__term.pid, signal.SIGTERM)
        self.remove(self.__term)
        self.__term.destroy()
        self.__init_state()
        self.ssh_connect()
        self.emit('reconnect')

    def __on_child_exited(self, term):
        _logger.debug("disconnected")
        self.__cmd_exited = True
        self.__show_disconnected()

    def __show_disconnected(self):
        self.__msg.hide()
        if self.__global_connection_changed:
            return
        msgarea = self.__msgarea_mgr.new_from_text_and_icon(gtk.STOCK_INFO, _('Connection closed'),
                                                            buttons=[(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)])
        reconnect = self.__actions.get_action('Reconnect')
        msgarea.add_stock_button_with_text(reconnect.get_property('label'),
                                           reconnect.get_property('stock-id'), gtk.RESPONSE_ACCEPT)
        msgarea.connect('response', self.__on_msgarea_response)
        msgarea.show_all()

    def __on_msgarea_response(self, msgarea, respid):
        if respid == gtk.RESPONSE_ACCEPT:
            self.ssh_reconnect()
        else:
            self.emit('close')

    def has_close(self):
        return True

    def get_exited(self):
        return self.__cmd_exited

    def get_term(self):
        return self.__term

    def get_vte(self):
        return self.__term.get_vte()

    def get_title(self):
        return self.get_host()

    def get_pixbuf(self):
        return self.__favicon_pixbuf

    def get_host(self):
        return self.__host

    def get_user(self):
        return self.__user

    def get_port(self):
        return self.__port

    def get_options(self):
        return self.__sshopts

class SshWindow(VteWindow):
    def __init__(self, **kwargs):
        super(SshWindow, self).__init__(title='Secure Shell', icon_name='hotwire-openssh', **kwargs)

        self.__ui_string = """
<ui>
  <menubar name='Menubar'>
    <menu action='FileMenu'>
      <placeholder name='FileAdditions'>
        <menuitem action='NewConnection'/>
        <menuitem action='CopyConnection'/>
        <menuitem action='OpenSFTP'/>
        <separator/>
        <menuitem action='Reconnect'/>
        <menuitem action='ReconnectAll'/>
        <separator/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

        self._get_notebook().connect('switch-page', self.__on_page_switch)

        try:
            self.__cached_nm_connected = None
            self.__nm_proxy = dbus.SystemBus().get_object('org.freedesktop.NetworkManager', '/org/freedesktop/NetworkManager')
            self.__nm_proxy.connect_to_signal('StateChange', self.__on_nm_state_change)
            dbus.Interface(self.__nm_proxy, 'org.freedesktop.DBus.Properties').Get('org.freedesktop.NetworkManager',
                                                                                   'State', reply_handler=self.__on_nm_state_change,
                                                                                   error_handler=self.__on_dbus_error)
        except dbus.DBusException, e:
            _logger.debug("Couldn't find NetworkManager")
            self.__nm_proxy = None

        self.__status_hbox = gtk.HBox()
        self._get_vbox().pack_start(self.__status_hbox, expand=False)
        self.__statusbar = gtk.Statusbar()
        self.__status_hbox.pack_start(self.__statusbar, expand=True)
        self.__statusbar_ctx = self.__statusbar.get_context_id("HotSSH")

        self.__in_reconnect = False
        self.__idle_stop_monitoring_id = 0

        self.connect("notify::is-active", self.__on_is_active_changed)
        _hostmonitor.connect('host-status', self.__on_host_status)

        self.__connhistory = get_history()
        self.__local_avahi = get_local_avahi()

        self.__merge_ssh_ui()

    def __add_to_history(self, args, inituser=None, inithost=None):
        user = inituser
        host = inithost
        need_arg = False
        options = []

        for arg in args:
            if arg.startswith('-') and SshTerminalWidget.ssh_opt_with_args.find(arg[1:]) >= 0:
                options.append(arg)
                if SshTerminalWidget.ssh_opt_with_args.find(arg[1:]) >= 0:
                    need_arg = True
                continue

            if need_arg:
                options.append(arg)
            elif host is None:
                if arg.find('@') >= 0:
                    (user, host) = arg.split('@', 1)
                else:
                    host = arg
            need_arg = False

        self.__connhistory.add_connection(user, host, options)

    def remote_new_tab(self, args, cwd):
        return self.new_tab(args, cwd, exit_on_cancel=False)

    def new_tab(self, args, cwd, exit_on_cancel=True, **kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            self.open_connection_dialog(exit_on_cancel=exit_on_cancel)
        else:
            self.__add_to_history(args, **kwargs)
            term = SshTerminalWidget(args=args, cwd=cwd, actions=self.__action_group, **kwargs)
            self.append_widget(term)

    def append_widget(self, w):
        super(SshWindow, self).append_widget(w)
        w.connect('status-changed', self.__on_widget_status_changed)
        w.connect('reconnect', self.__on_widget_reconnect)
        self.__sync_status_display()

    def __on_widget_status_changed(self, w):
        self.__sync_status_display()

    def __on_widget_reconnect(self, changed_widget):
        if self.__in_reconnect:
            return
        self.__in_reconnect = True
        changed_user = changed_widget.get_user()
        changed_host = changed_widget.get_host()
        changed_userhost = userhost_pair_to_string(changed_user, changed_host)
        _logger.debug("reconnecting all widgets for host %s", changed_userhost)
        for widget in self._get_notebook().get_children():
            if changed_widget is widget:
                continue
            host = widget.get_host()
            user = widget.get_user()
            userhost = userhost_pair_to_string(user, host)
            if userhost != changed_userhost:
                continue
            widget.ssh_reconnect()
        self.__in_reconnect = False

    def __sync_status_display(self):
        notebook = self._get_notebook()
        pn = notebook.get_current_page()
        if pn < 0:
            return
        widget = notebook.get_nth_page(pn)
        if widget.connecting_state:
            text = _('Connecting')
        elif widget.connected is True:
            text = _('Connected; %.2fs latency; %s') % (widget.latency, widget.status_line)
        elif widget.connected is False:
            text = _('Connection timed out')
        elif widget.connected is None:
            text = _('Checking connection')
        #if len(widget.ssh_options) > 1:
        #    text += _('; Options: ') + (' '.join(map(gobject.markup_escape_text, widget.ssh_options)))
        id = self.__statusbar.push(self.__statusbar_ctx, text)

    def __on_dbus_error(self, *args, **kwargs):
        _logger.error("DBus error: %r %r", args, kwargs)

    @log_except(_logger)
    def __on_nm_state_change(self, *args):
        if len(args) > 1:
            (_, curstate) = args
        else:
            curstate = args[0]
        _logger.debug("nm state: %s", curstate)
        connected = curstate == 3
        if (self.__cached_nm_connected is not None) and \
            (not self.__cached_nm_connected) and connected:
            self.__do_network_change_notify()
        self.__cached_nm_connected = connected
        for child in self._get_notebook():
            child.set_global_connection_changed(True)

    def __do_network_change_notify(self):
        mgr = self._get_msgarea_mgr()
        msgarea = mgr.new_from_text_and_icon(gtk.STOCK_INFO, _('Network connection changed'),
                                                            buttons=[(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)])
        reconnect = self.__action_group.get_action('ReconnectAll')
        msgarea.add_stock_button_with_text(reconnect.get_property('label'),
                                           reconnect.get_property('stock-id'), gtk.RESPONSE_ACCEPT)
        msgarea.connect('response', self.__on_msgarea_response)
        msgarea.show_all()

    def __on_msgarea_response(self, msgarea, respid):
        mgr = self._get_msgarea_mgr()
        mgr.clear()
        if respid == gtk.RESPONSE_ACCEPT:
            for child in self._get_notebook():
                child.set_global_connection_changed(False)
            reconnect = self.__action_group.get_action('ReconnectAll')
            reconnect.activate()

    @log_except(_logger)
    def __on_host_status(self, hostmon, userhost, connected, latency, status_output):
        _logger.debug("got host status host=%s conn=%s latency=%s", userhost, connected, latency)
        for widget in self._get_notebook().get_children():
            child_userhost = userhost_pair_to_string(widget.get_user(), widget.get_host())
            if child_userhost != userhost:
                continue
            widget.set_status(connected, latency, status_output)

    @log_except(_logger)
    def __on_is_active_changed(self, *args):
        isactive = self.get_property('is-active')
        if isactive:
            self.__start_monitoring()
            if self.__idle_stop_monitoring_id > 0:
                gobject.source_remove(self.__idle_stop_monitoring_id)
                self.__idle_stop_monitoring_id = 0
        elif self.__idle_stop_monitoring_id == 0:
            self.__idle_stop_monitoring_id = gobject.timeout_add(8000, self.__idle_stop_monitoring)

    def __idle_stop_monitoring(self):
        self.__idle_stop_monitoring_id = 0
        self.__stop_monitoring()

    @log_except(_logger)
    def __on_page_switch(self, n, p, pn):
        # Becuase of the way get_current_page() works in this signal handler, this
        # will actually disable monitoring for the previous tab, and enable it
        # for the new current one.
        self.__sync_monitoring(new_pn=pn)
        self.__sync_status_display()

    def __get_userhost_for_pn(self, pn):
        notebook = self._get_notebook()
        if pn >= 0:
            widget = notebook.get_nth_page(pn)
            return userhost_pair_to_string(widget.get_user(), widget.get_host())
        return None

    def __stop_monitoring(self):
        notebook = self._get_notebook()
        pn = notebook.get_current_page()
        if pn >= 0:
            prev_widget = notebook.get_nth_page(pn)
            prev_user = prev_widget.get_user()
            prev_host = prev_widget.get_host()
            _hostmonitor.stop_monitor(prev_user, prev_host)
            prev_widget.set_status(None, None)

    def __start_monitoring(self, pn=None):
        notebook = self._get_notebook()
        if pn is not None:
            pagenum = pn
        else:
            pagenum = notebook.get_current_page()
        widget = notebook.get_nth_page(pagenum)
        _hostmonitor.start_monitor(widget.get_user(), widget.get_host())

    def __sync_monitoring(self, new_pn):
        prev_userhost = self.__get_userhost_for_pn(self._get_notebook().get_current_page())
        new_userhost = self.__get_userhost_for_pn(new_pn)
        if prev_userhost != new_userhost:
            self.__stop_monitoring()
            self.__start_monitoring(pn=new_pn)

    def __merge_ssh_ui(self):
        self.__using_accels = True
        self.__actions = actions = [
            ('NewConnection', gtk.STOCK_NEW, _('Connect to server'), '<control><shift>O',
             _('Open a new Secure Shell connection'), self.__new_connection_cb),
            ('CopyConnection', gtk.STOCK_JUMP_TO, _('New tab for connection'), '<control><shift>T',
             _('Open a new tab for the same remote computer'), self.__copy_connection_cb),
            ('OpenSFTP', gtk.STOCK_OPEN, _('Open SFTP'), '<control><shift>S',
             _('Open a SFTP connection'), self.__open_sftp_cb),
            ('ConnectionMenu', None, _('Connection')),
            ('Reconnect', gtk.STOCK_CONNECT, _('_Reconnect'), '<control><shift>R', _('Reset connection to server'), self.__reconnect_cb),
            ('ReconnectAll', gtk.STOCK_CONNECT, _('R_econnect All'), None, _('Reset all connections'), self.__reconnect_all_cb),
            ]
        self.__action_group = self._merge_ui(self.__actions, self.__ui_string)

    @log_except(_logger)
    def __copy_connection_cb(self, action):
        notebook = self._get_notebook()
        widget = notebook.get_nth_page(notebook.get_current_page())
        user = widget.get_user()
        host = widget.get_host()
        opts = widget.get_options()
        self.new_tab(widget.get_options(), None, inituser=user, inithost=host)

    def open_connection_dialog(self, exit_on_cancel=False):
        win = ConnectDialog(parent=self, history=self.__connhistory, local_avahi=self.__local_avahi)
        sshargs = win.run_get_cmd()
        if not sshargs:
            # We get here when called with no arguments, and we're the main instance.
            if exit_on_cancel:
                sys.exit(0)
            return
        self.new_tab(sshargs, None, exit_on_cancel=exit_on_cancel)

    @log_except(_logger)
    def __new_connection_cb(self, action):
        self.open_connection_dialog()

    @log_except(_logger)
    def __open_sftp_cb(self, action):
        notebook = self._get_notebook()
        widget = notebook.get_nth_page(notebook.get_current_page())
        host = widget.get_host()
        port = widget.get_port()
        if port:
            subprocess.Popen(['nautilus', 'sftp://%s:%s' % (host,port)])
        else:
            subprocess.Popen(['nautilus', 'sftp://%s' % (host,)])

    @log_except(_logger)
    def __reconnect_cb(self, a):
        notebook = self._get_notebook()
        widget = notebook.get_nth_page(notebook.get_current_page())
        widget.ssh_reconnect()

    @log_except(_logger)
    def __reconnect_all_cb(self, a):
        for widget in self._get_notebook().get_children():
            widget.ssh_reconnect()

    def _do_about(self):
        dlg = HotSshAboutDialog()
        dlg.run()
        dlg.destroy()

class SshApp(VteApp):
    def __init__(self):
        super(SshApp, self).__init__(SshWindow)
        self.__old_sessionpath = os.path.expanduser('~/.hotwire/hotssh.session')
        self.__sessionpath = os.path.expanduser('~/.hotssh/hotssh-session.xml')
        self.__connhistory = get_history()
        self.__local_avahi = SshAvahiMonitor()

    @staticmethod
    def get_name():
        return 'HotSSH'

    def on_shutdown(self, factory):
        super(SshApp, self).on_shutdown(factory)
        cp = get_controlpath(create=False)
        if cp is not None:
            try:
                _logger.debug("removing %s", cp)
                shutil.rmtree(cp)
            except:
                pass

    def offer_load_session(self):
        savedsession = self._parse_saved_session()
        alltabs_count = 0
        allhosts = set()
        if savedsession:
            for window in savedsession:
                for connection in window:
                    allhosts.add(connection['userhost'])
                    alltabs_count += 1
        if len(allhosts) > 0:
            dlg = gtk.MessageDialog(parent=None, flags=0, type=gtk.MESSAGE_QUESTION,
                                    buttons=gtk.BUTTONS_CANCEL,
                                    message_format=_("Restore saved session?"))
            button = dlg.add_button(_('_Reconnect'), gtk.RESPONSE_ACCEPT)
            button.set_property('image', gtk.image_new_from_stock('gtk-connect', gtk.ICON_SIZE_BUTTON))
            dlg.set_default_response(gtk.RESPONSE_ACCEPT)
            allhosts_count = len(allhosts)
            # Translators: %d is the number of hosts we are about to reconnect to.
            text = gettext.ngettext('Reconnect to %d host' % (allhosts_count,),
                                    'Reconnect to %d hosts' % (allhosts_count,), len(allhosts))
            dlg.format_secondary_markup(text)

            #ls = gtk.ListStore(str)
            #gv = gtk.TreeView(ls)
            #colidx = gv.insert_column_with_attributes(-1, _('Connection'),
            #                                          gtk.CellRendererText(),
            #                                          text=0)
            #for host in allhosts:
            #    ls.append((host,))
            #dlg.add(gv)

            resp = dlg.run()
            dlg.destroy()
            if resp == gtk.RESPONSE_ACCEPT:
                self._load_session(savedsession)
                return
        w = self.get_factory().create_initial_window()
        w.new_tab([], os.getcwd())
        w.show_all()
        w.present()

    def _load_session(self, session):
        factory = self.get_factory()
        for window in session:
            window_impl = factory.create_window()
            for connection in window:
                args = [connection['userhost']]
                if 'options' in connection:
                    args.extend(connection['options'])
                widget = window_impl.new_tab(args, os.getcwd())
            window_impl.show_all()

    #override
    @log_except(_logger)
    def _parse_saved_session(self):
        factory = self.get_factory()
        try:
            f = open(self.__sessionpath)
        except:
            try:
                f = open(self.__old_sessionpath)
            except:
                return None
        doc = xml.dom.minidom.parse(f)
        saved_windows = []
        current_widget = None
        for window_child in doc.documentElement.childNodes:
            if not (window_child.nodeType == window_child.ELEMENT_NODE and window_child.nodeName == 'window'):
                continue
            connections = []
            for child in window_child.childNodes:
                if not (child.nodeType == child.ELEMENT_NODE and child.nodeName == 'connection'):
                    continue
                user = child.getAttribute('user')
                host = child.getAttribute('host')
                iscurrent = child.getAttribute('current')
                options = []
                for options_elt in child.childNodes:
                    if not (options_elt.nodeType == child.ELEMENT_NODE and options_elt.nodeName == 'options'):
                        continue
                    for option_elt in child.childNodes:
                        if not (option_elt.nodeType == child.ELEMENT_NODE and options_elt.nodeName == 'option'):
                            continue
                        options.append(option.firstChild.nodeValue)
                userhost = userhost_pair_to_string(user, host)
                kwargs = {'userhost': userhost}
                if len(options) > 0:
                    kwargs['options'] = options
                if iscurrent:
                    kwargs['current'] = True
                connections.append(kwargs)
            saved_windows.append(connections)
        return saved_windows

    #override
    @log_except(_logger)
    def save_session(self):
        _logger.debug("doing session save")
        tempf_path = tempfile.mktemp('.xml.tmp', 'hotssh-session', os.path.dirname(self.__sessionpath))
        f = open(tempf_path, 'w')
        state = []
        doc = xml.dom.minidom.getDOMImplementation().createDocument(None, "session", None)
        root = doc.documentElement
        factory = self.get_factory()
        for window in factory._get_windows():
            notebook = window._get_notebook()
            current = notebook.get_nth_page(notebook.get_current_page())
            window_node = doc.createElement('window')
            root.appendChild(window_node)
            for widget in notebook:
                connection = doc.createElement('connection')
                window_node.appendChild(connection)
                user = widget.get_user()
                if user is not None:
                    connection.setAttribute('user', user)
                host = widget.get_host()
                if host is not None:
                    connection.setAttribute('host', host)
                if current == widget:
                    connection.setAttribute('current', 'true')
                options = widget.get_options()
                if options:
                    options_elt = doc.createElement('options')
                    connection.appendChild(options_elt)
                    for option in options:
                        opt = doc.createElement('option')
                        options_elt.appendChild(opt)
                        opt.appendChild(doc.createTextNode(option))
        f.write(doc.toprettyxml())
        f.close()
        os.rename(tempf_path, self.__sessionpath)



#! /usr/bin/env python
# encoding: utf-8

import os,sys
import Build,TaskGen,Utils,Options,Logs
import pproc
class unit_test(object):
	def __init__(self):
		self.returncode_ok=0
		self.num_tests_ok=0
		self.num_tests_failed=0
		self.num_tests_err=0
		self.total_num_tests=0
		self.max_label_length=0
		self.unit_tests={}
		self.unit_test_results={}
		self.unit_test_erroneous={}
		self.change_to_testfile_dir=False
		self.want_to_see_test_output=False
		self.want_to_see_test_error=False
		self.run_if_waf_does='check'
	def run(self):
		self.num_tests_ok=0
		self.num_tests_failed=0
		self.num_tests_err=0
		self.total_num_tests=0
		self.max_label_length=0
		self.unit_tests={}
		self.unit_test_results={}
		self.unit_test_erroneous={}
		if not Options.commands[self.run_if_waf_does]:return
		for obj in Build.bld.all_task_gen:
			if not hasattr(obj,'unit_test'):continue
			unit_test=getattr(obj,'unit_test')
			if not unit_test:continue
			try:
				if'program'in obj.features:
					output=obj.path
					filename=os.path.join(output.abspath(obj.env),obj.target)
					srcdir=output.abspath()
					label=os.path.join(output.bldpath(obj.env),obj.target)
					self.max_label_length=max(self.max_label_length,len(label))
					self.unit_tests[label]=(filename,srcdir)
			except KeyError:
				pass
		self.total_num_tests=len(self.unit_tests)
		Utils.pprint('GREEN','Running the unit tests')
		count=0
		result=1
		for label,file_and_src in self.unit_tests.iteritems():
			filename=file_and_src[0]
			srcdir=file_and_src[1]
			count+=1
			line=Build.bld.progress_line(count,self.total_num_tests,Logs.colors.GREEN,Logs.colors.NORMAL)
			if Options.options.progress_bar and line:
				sys.stdout.write(line)
				sys.stdout.flush()
			try:
				kwargs={}
				if self.change_to_testfile_dir:
					kwargs['cwd']=srcdir
				if not self.want_to_see_test_output:
					kwargs['stdout']=pproc.PIPE
				if not self.want_to_see_test_error:
					kwargs['stderr']=pproc.PIPE
				pp=pproc.Popen(filename,**kwargs)
				pp.wait()
				result=int(pp.returncode==self.returncode_ok)
				if result:
					self.num_tests_ok+=1
				else:
					self.num_tests_failed+=1
				self.unit_test_results[label]=result
				self.unit_test_erroneous[label]=0
			except OSError:
				self.unit_test_erroneous[label]=1
				self.num_tests_err+=1
			except KeyboardInterrupt:
				pass
		if Options.options.progress_bar:sys.stdout.write(Logs.colors.cursor_on)
	def print_results(self):
		if not Options.commands[self.run_if_waf_does]:return
		p=Utils.pprint
		if self.total_num_tests==0:
			p('YELLOW','No unit tests present')
			return
		p('GREEN','Running unit tests')
		print
		for label,filename in self.unit_tests.iteritems():
			err=0
			result=0
			try:err=self.unit_test_erroneous[label]
			except KeyError:pass
			try:result=self.unit_test_results[label]
			except KeyError:pass
			n=self.max_label_length-len(label)
			if err:n+=4
			elif result:n+=7
			else:n+=3
			line='%s %s'%(label,'.'*n)
			print line,
			if err:p('RED','ERROR')
			elif result:p('GREEN','OK')
			else:p('YELLOW','FAILED')
		percentage_ok=float(self.num_tests_ok)/float(self.total_num_tests)*100.0
		percentage_failed=float(self.num_tests_failed)/float(self.total_num_tests)*100.0
		percentage_erroneous=float(self.num_tests_err)/float(self.total_num_tests)*100.0
		print'''
Successful tests:      %i (%.1f%%)
Failed tests:          %i (%.1f%%)
Erroneous tests:       %i (%.1f%%)

Total number of tests: %i
'''%(self.num_tests_ok,percentage_ok,self.num_tests_failed,percentage_failed,self.num_tests_err,percentage_erroneous,self.total_num_tests)
		p('GREEN','Unit tests finished')




# This file is part of the Hotwire Shell project API.

# Copyright (C) 2007 Colin Walters <walters@verbum.org>

# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal 
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
# of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all 
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE X CONSORTIUM BE 
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR 
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys, logging, logging.config, StringIO

def log_except(logger=None, text=''):
    def annotate(func):
        def _exec_cb(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                log_target = logger or logging
                log_target.exception('Exception in callback%s', text and (': '+text) or '')
        return _exec_cb
    return annotate

def init(default_level, debug_modules, prefix=None):
    rootlog = logging.getLogger() 
    fmt = logging.Formatter("%(asctime)s [%(thread)d] %(name)s %(levelname)s %(message)s",
                            "%H:%M:%S")
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setFormatter(fmt)
    
    rootlog.setLevel(default_level)
    rootlog.addHandler(stderr_handler)
    for logger in [logging.getLogger(prefix+x) for x in debug_modules]:
        logger.setLevel(logging.DEBUG)

    logging.debug("Initialized logging")



# This file is part of the Hotwire Shell user interface.
#   
# Copyright (C) 2007 Colin Walters <walters@verbum.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os,sys,platform,logging,getopt
import locale,threading,subprocess,time
import signal

import gtk,gobject,pango
import dbus,dbus.glib,dbus.service

from hotssh.hotvte.vteterm import VteTerminalWidget
from hotssh.hotlib_ui.msgarea import MsgAreaController
from hotssh.hotlib_ui.quickfind import QuickFindWindow

_logger = logging.getLogger("hotvte.VteWindow")

class QuickSwitchTabWindow(QuickFindWindow):
    def __init__(self, vtewin):
        self.__vtewin = vtewin        
        super(QuickSwitchTabWindow, self).__init__(_('Tab Search'),
                                                   parent=vtewin)
        
    def _do_search(self, text):
        for widget in self.__vtewin.get_tabs():
            title = widget.get_title()
            markup = self._markup_search(title, text)
            if markup is not None:
                yield (title, markup, None)

class TabbedVteWidget(VteTerminalWidget):
    __gsignals__ = {
        "metadata-changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
    }    

    def __init__(self, cmd=None, *args, **kwargs):
        super(TabbedVteWidget, self).__init__(cmd=cmd, *args, **kwargs)
        self.__title = ' '.join(cmd)
        
    def get_title(self):
        return self.__title

    def get_pixbuf(self):
        return None

class VteWindow(gtk.Window):
    __gsignals__ = {
        "window-close" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
        "shutdown" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
    }    

    ascii_nums = [long(x+ord('0')) for x in xrange(10)]    
    def __init__(self, factory=None, title=None, icon_name=None, **kwargs):
        super(VteWindow, self).__init__()
        
        self.__factory = factory
        
        self.__idle_save_session_id = 0
        
        self.__old_char_height = 0
        self.__old_char_width = 0
        self.__old_geom_widget = None
        
        self.__vbox = vbox = gtk.VBox()
        self.add(vbox)
        self.__ui_string = """
<ui>
  <menubar name='Menubar'>
    <menu action='FileMenu'>
      <placeholder name='FileAdditions'/>
      <separator/>
      <menuitem action='TabSearch'/>      
      <menuitem action='DetachTab'/>
      <separator/>
      <menuitem action='Close'/>
      <separator/>
      <menuitem action='Quit'/>
    </menu>
    <menu action='EditMenu'>
      <menuitem action='Copy'/>
      <menuitem action='Paste'/>
      <separator/>
      <placeholder name='EditAdditions'/>      
    </menu>
    <menu action='ViewMenu'>
      <separator/>
      <placeholder name='ViewAdditions'/>        
    </menu>
    <placeholder name='TermAppAdditions'/>
    <menu action='HelpMenu'>
      <menuitem action='About'/>
    </menu>
  </menubar>
</ui>
"""       
        self.__create_ui()
        vbox.pack_start(self.__ui.get_widget('/Menubar'), expand=False)
        
        self.__msgarea_mgr = MsgAreaController()
        vbox.pack_start(self.__msgarea_mgr, expand=False)        
        
        self.connect("key-press-event", self.__on_keypress)
                
        self.__title = title
        self.set_title(title)
        self.set_default_size(720, 540)
        if os.getenv('HOTWIRE_UNINSTALLED'):
            # For some reason set_icon() started failing...do it manually.
            iinf = gtk.icon_theme_get_default().lookup_icon(icon_name, 24, 0)
            if iinf:
                fn = iinf.get_filename()
                self.set_icon_from_file(fn)
        else:
            self.set_icon_name(icon_name)
        self.connect("delete-event", lambda w, e: False)
        self.__tips = gtk.Tooltips()
        self.__notebook = gtk.Notebook()
        vbox.pack_start(self.__notebook)
        self.__notebook.connect('switch-page', self.__on_page_switch)
        self.__notebook.set_scrollable(True)
        self.__notebook.show()
        
        self.__tabs_visible = self.__notebook.get_show_tabs()
        self.__queue_session_save()
        
    def __queue_session_save(self):
        self.__factory.queue_save_session()        
        
    def new_tab(self, args, cwd):
        widget = TabbedVteWidget(cmd=args, cwd=cwd)
        return self.append_widget(widget)
        
    def remote_new_tab(self, args, cwd):
        return self.new_tab(args, cwd)

    def __sync_metadata(self, widget):
        title = widget.get_title()
        label = widget.get_data('tab-label')
        label.set_text(title)
        self.__tips.set_tip(label, title)
        img = widget.get_data('tab-image')
        pixbuf_or_path = widget.get_pixbuf()
        if isinstance(pixbuf_or_path,basestring):
            pixbuf = gtk.gdk.pixbuf_new_from_file(pixbuf_or_path)
        else:
            pixbuf = pixbuf_or_path
        img.set_from_pixbuf(pixbuf)

    def append_widget(self, term):
        idx = self.__notebook.append_page(term)
        term.connect('metadata-changed', self.__sync_metadata)
        term.get_vte().connect('selection-changed', self.__sync_selection_sensitive)
        term.get_term().set_copy_paste_actions(self.__ag.get_action('Copy'), self.__ag.get_action('Paste'))
        if hasattr(term, 'has_close'):
            has_close = term.has_close()
        else:
            has_close = False
        if has_close:
            term.connect('close', self.__on_widget_close)
        if hasattr(self.__notebook, 'set_tab_reorderable'):
            self.__notebook.set_tab_reorderable(term, True)
        (label, icon) = self.__add_widget_title(term)
        self.__sync_metadata(term)
        term.show_all()
        self.__notebook.set_current_page(idx)
        term.get_vte().grab_focus()
        return term
        
    def _get_notebook(self):
        return self.__notebook
    
    def _get_vbox(self):
        return self.__vbox
    
    def _get_msgarea_mgr(self):
        return self.__msgarea_mgr

    def __on_page_switch(self, n, p, pn):
        _logger.debug("got page switch, pn=%d", pn)
        widget = self.__notebook.get_nth_page(pn)
        term = widget.get_vte()
        (cw, ch, (xp, yp)) = (term.get_char_width(), term.get_char_height(), term.get_padding())
        if not (cw == self.__old_char_width and ch == self.__old_char_height and widget == self.__old_geom_widget):
            _logger.debug("resetting geometry on %s %s %s => %s %s", widget, self.__old_char_width, self.__old_char_height, cw, ch)
            kwargs = {'base_width':xp,
                      'base_height':yp,
                      'width_inc':cw,
                      'height_inc':ch,
                      'min_width':xp+cw*4,
                      'min_height':yp+ch*2}
            _logger.debug("setting geom hints: %s", kwargs)
            self.__geom_hints = kwargs
            self.set_geometry_hints(widget, **kwargs)
            self.__old_char_width = cw
            self.__old_char_height = ch
            self.__old_geom_widget = widget
            
        self.__sync_selection_sensitive()
        self.set_title('%s - %s' % (widget.get_title(),self.__title))
        self.__queue_session_save()
        
    def __on_keypress(self, s2, e):
        if e.keyval == gtk.gdk.keyval_from_name('Page_Up') and \
             e.state & gtk.gdk.CONTROL_MASK:
            idx = self.__notebook.get_current_page() 
            self.__notebook.set_current_page(idx-1)
            return True
        elif e.keyval == gtk.gdk.keyval_from_name('Page_Down') and \
             e.state & gtk.gdk.CONTROL_MASK:
            idx = self.__notebook.get_current_page() 
            self.__notebook.set_current_page(idx+1)
            return True
        elif e.keyval in VteWindow.ascii_nums and \
             e.state & gtk.gdk.MOD1_MASK:
            self.__notebook.set_current_page(e.keyval-ord('0')-1) #extra -1 because tabs are 0-indexed
            return True
        elif e.keyval == gtk.gdk.keyval_from_name('Return'):
            widget = self.__notebook.get_nth_page(self.__notebook.get_current_page())        
            if widget.get_exited():
                self.__close_tab(widget)
                return True
        return False        
        
    def __create_ui(self):
        self.__using_accels = True
        self.__ag = ag = gtk.ActionGroup('WindowActions')
        self.__actions = actions = [
            ('FileMenu', None, _('File')),
            ('DetachTab', None, _('_Detach Tab'), '<control><shift>D', 'Move tab into new window', self.__detach_cb),
            ('TabSearch', None, '_Search Tabs', '<control><shift>L', 'Search across tab names', self.__quickswitch_tab_cb),
            ('Close', gtk.STOCK_CLOSE, _('_Close'), '<control><shift>W',
             'Close the current tab', self.__close_cb),
            ('Quit', gtk.STOCK_QUIT, _('_Quit'), '<control><shift>Q',
             'Quit application', self.__quit_cb),
            ('EditMenu', None, _('Edit')),
            ('Copy', gtk.STOCK_COPY, _('_Copy'), '<control><shift>C', 'Copy selected text', self.__copy_cb),
            ('Paste', gtk.STOCK_PASTE, _('_Paste'), '<control><shift>V', 'Paste text', self.__paste_cb),                   
            ('ViewMenu', None, _('View')),
            ('HelpMenu', None, _('Help')),
            ('About', gtk.STOCK_ABOUT, _('_About'), None, 'About this application', self.__help_about_cb),
            ]
        ag.add_actions(actions)
        self.__ui = gtk.UIManager()
        self.__ui.insert_action_group(ag, 0)
        self.__ui.add_ui_from_string(self.__ui_string)
        self.add_accel_group(self.__ui.get_accel_group())
        
    def _merge_ui(self, actions, uistr):
        self.__ag.add_actions(actions)
        self.__ui_merge_id = self.__ui.add_ui_from_string(uistr)
        return self.__ag      
    
    def _get_ui(self):
        return self.__ui   

    def __add_widget_title(self, w):
        hbox = gtk.HBox()
        piximg = gtk.Image()
        w.set_data('tab-image', piximg)
        hbox.pack_start(piximg, expand=False, padding=6)
        label = gtk.Label('<notitle>')
        label.set_selectable(False)
        label.set_ellipsize(pango.ELLIPSIZE_END)
        hbox.pack_start(label, expand=True)

        close = gtk.Button()
        close.set_focus_on_click(False)
        close.set_relief(gtk.RELIEF_NONE)
        close.set_name('hotwire-tab-close')
        img = gtk.Image()
        img.set_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
        close.add(img)
        close.connect('clicked', lambda b: self.__close_tab(w))        
        (width, height) = gtk.icon_size_lookup_for_settings(label.get_settings(), gtk.ICON_SIZE_MENU)
        close.set_size_request(width + 2, height + 2)
        hbox.pack_start(close, expand=False)
        hbox.show_all()
        self.__notebook.set_tab_label(w, hbox)
        w.set_data('tab-label', label)
        self.__notebook.set_tab_label_packing(w, True, True, gtk.PACK_START)
        self.__sync_tabs_visible()
        return (label, piximg)
    
    def __close_tab(self, w):
        self.__remove_page_widget(w)
        w.destroy()
        
    def __close_cb(self, action):
        self.__remove_page_widget(self.__notebook.get_nth_page(self.__notebook.get_current_page()))

    def __quit_cb(self, action):
        self.emit('shutdown')
        
    def __on_widget_close(self, widget):
        self.__remove_page_widget(widget)

    def __help_about_cb(self, action):
        self._do_about()
        
    def _do_about(self):
        raise NotImplementedError()

    def get_tabs(self):
        return self.__notebook.get_children()
        
    def __quickswitch_tab_cb(self, action):
        w = QuickSwitchTabWindow(self)
        tabtitle = w.run_get_value()
        if tabtitle is None:
            return
        _logger.debug("got switch title: %r", tabtitle)
        for child in self.__notebook.get_children():
            if child.get_title() == tabtitle:
                self.__notebook.set_current_page(self.__notebook.page_num(child))
                break
        
    def __sync_tabs_visible(self):
        multitab = self.__notebook.get_n_pages() > 1
        self.__ag.get_action('DetachTab').set_sensitive(multitab)
        self.__ag.get_action('TabSearch').set_sensitive(multitab)        
        self.__notebook.set_show_tabs(multitab)        
        
    def __remove_page_widget(self, w):
        idx = self.__notebook.page_num(w)
        _logger.debug("tab closed, current: %d", idx)
        self.__notebook.remove_page(idx)
        self.__sync_tabs_visible()
        if self.__notebook.get_n_pages() == 0:
            self.destroy()      

    def __sync_selection_sensitive(self, *args):
        have_selection = self.__notebook.get_nth_page(self.__notebook.get_current_page()).get_vte().get_has_selection()
        self.__ag.get_action('Copy').set_sensitive(have_selection)

    def __copy_cb(self, a):
        _logger.debug("doing copy")
        widget = self.__notebook.get_nth_page(self.__notebook.get_current_page())        
        widget.get_vte().copy_clipboard()        

    def __paste_cb(self, a):
        _logger.debug("doing paste")
        widget = self.__notebook.get_nth_page(self.__notebook.get_current_page())
        widget.get_vte().paste_clipboard()

    def __detach_cb(self, a):
        widget = self.__notebook.get_nth_page(self.__notebook.get_current_page())
        self.__remove_page_widget(widget)               
        win = self.__factory.create_window()
        win.append_widget(widget)
        win.show_all()
                
_factory = None
class VteWindowFactory(gobject.GObject):
    __gsignals__ = {
        "shutdown" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
    }    
    def __init__(self, klass, window_args, app):
        super(VteWindowFactory, self).__init__()
        self.__windows = set()
        self.__klass = klass
        self.__app = app
        self.__window_args = window_args
        self.__sticky_keywords = {'subtitle': ''}
        self.__sticky_keywords.update(window_args)
        self.__recentwindow = None
        self.__idle_save_session_id = 0        
        
    @staticmethod
    def getInstance():
        global _factory
        if _factory is None:
            _factory = VteWindowFactory()
        return _factory

    def offer_load_session(self):
        dlg = Load

    def create_initial_window(self, *args, **kwargs):
        win = self.create_window(is_initial=True, *args, **kwargs)
        self.__recentwindow = win
        return win

    def create_window(self, is_initial=False, *args, **kwargs):
        _logger.debug("creating window")
        if is_initial:
            for k,v in kwargs.iteritems():
                if self.__sticky_keywords.has_key(k):
                    self.__sticky_keywords[k] = v
        for k,v in self.__sticky_keywords.iteritems():
            if not kwargs.has_key(k):
                kwargs[k] = v
        win = self.__klass(factory=self, is_initial=is_initial, **kwargs)
        win.connect('notify::is-active', self.__on_window_active)
        win.connect('destroy', self.__on_win_destroy)
        win.connect('shutdown', self.__on_win_shutdown)
        self.__windows.add(win)
        return win
    
    def remote_new_tab(self, cmd, cwd):
        self.__recentwindow.remote_new_tab(cmd, cwd)
        return self.__recentwindow
    
    def queue_save_session(self):      
        if self.__idle_save_session_id == 0:
            self.__idle_save_session_id = gobject.timeout_add(5*1000, self.__idle_save_session)

    def __save_session_now(self):
        if self.__idle_save_session_id > 0:
            gobject.source_remove(self.__idle_save_session_id)
        self.__idle_save_session()
            
    def __idle_save_session(self):
        self.__idle_save_session_id = 0
        self.__app.save_session()
    
    def _get_windows(self):
        return self.__windows
    
    def __on_window_active(self, win, *args):
        active = win.get_property('is-active')
        if active:
            self.__recentwindow = win

    def __on_win_destroy(self, win):
        _logger.debug("got window destroy")
        self.__save_session_now()
        self.__windows.remove(win)
        win.get_child().destroy()
        if len(self.__windows) == 0:
            self.emit('shutdown')
            gtk.main_quit()

    def __on_win_shutdown(self, win):
        self.emit('shutdown')
        gtk.main_quit()

class UiProxy(dbus.service.Object):
    def __init__(self, factory, bus_name, ui_iface, ui_opath):
        super(UiProxy, self).__init__(dbus.service.BusName(bus_name, bus=dbus.SessionBus()), ui_opath)
        self.__winfactory = factory
        # This is a disturbing hack.  But it works.
        def RunCommand(self, timestamp, istab, cmd, cwd):
            _logger.debug("Handling RunCommand method invocation ts=%r cmd=%r cwd=%r)", timestamp, cmd, cwd)
            if istab:
                curwin = self.__winfactory.remote_new_tab(cmd, cwd)
            else:
                raise NotImplementedError('can only create new tabs')
            timestamp = long(timestamp)+1
            if timestamp > 0:
                _logger.debug("presenting with timestamp %r", timestamp)
                curwin.present_with_time(timestamp)
            else:
                curwin.present()
        setattr(UiProxy, 'RunCommand', dbus.service.method(ui_iface, in_signature='ubass')(RunCommand))                

class VteRemoteControl(object):
    def __init__(self, name, bus_name=None, ui_opath=None, ui_iface=None):
        super(VteRemoteControl, self).__init__()
        self.__bus_name = bus_name or ('org.hotwireshell.HotVTE.' + name)
        self.__ui_opath = ui_opath or ('/hotvte/' + name + '/ui')
        self.__ui_iface = ui_iface or (self.__bus_name + '.Ui')
        
    def __parse_startup_id(self):
        startup_time = None
        try:
            startup_id_env = os.environ['DESKTOP_STARTUP_ID']
        except KeyError, e:
            startup_id_env = None
        if startup_id_env:
            idx = startup_id_env.find('_TIME')
            if idx > 0:
                idx += 5
                startup_time = int(startup_id_env[idx:])
        return startup_time        
        
    def single_instance(self, replace=False):
        proxy = dbus.SessionBus().get_object('org.freedesktop.DBus', '/org/freedesktop/DBus')
        flags = 1 | 4 # allow replacement | do not queue
        if replace:
            flags = flags | 2 # replace existing
        _logger.debug("Requesting D-BUS name %s on session bus", self.__bus_name)            
        if not proxy.RequestName(self.__bus_name, dbus.UInt32(flags)) in (1,4):
            inst = dbus.SessionBus().get_object(self.__bus_name, self.__ui_opath)
            inst_iface = dbus.Interface(inst, self.__ui_iface)
            _logger.debug("Sending RunCommand to existing instance")
            # TODO support choosing tab/window
            starttime = self.__parse_startup_id()
            inst_iface.RunCommand(dbus.UInt32(starttime or 0), True, dbus.Array(sys.argv[1:], signature="s"), os.getcwd())
            sys.exit(0)
            os._exit(0)
        
    def get_proxy(self, factory):
        return UiProxy(factory, self.__bus_name, self.__ui_iface, self.__ui_opath)
    
class VteApp(object): 
    def __init__(self, windowklass):
        super(VteApp, self).__init__()
        self.__windowklass = windowklass
        self.__factory = None
        
    @staticmethod
    def get_name():
        raise NotImplementedError()
        
    def get_remote(self):
        return VteRemoteControl(self.get_name())
        
    def get_factory(self):
        if self.__factory is None:
            self.__factory = VteWindowFactory(self.__windowklass, {}, self)
        return self.__factory
    
    def offer_load_session(self):
        pass
    
    def on_shutdown(self, factory):
        # We used to save the session here, but now that is done in
        # the window shutdown above
        pass
    
    def save_session(self):
        _logger.debug("noop session save")
        pass
    
class VteMain(object):
    def main(self, appklass):

        default_log_level = logging.WARNING
        if 'HOTVTE_DEBUG' in os.environ:
            default_log_level = logging.DEBUG
 
        import hotssh.hotlib.logutil
        mods = os.environ.get('HOTVTE_DEBUG_MODULES', '')
        if mods:
            mods = mods.split(',')
        else:
            mods = []
        hotssh.hotlib.logutil.init(default_log_level, mods, '')

        _logger.debug("logging initialized")

        locale.setlocale(locale.LC_ALL, '')
        import gettext
        gettext.install(appklass.get_name())      
    
        gobject.threads_init()
        
        app = appklass()
        remote = app.get_remote()
        remote.single_instance()

        def on_about_dialog_url(dialog, link):
            import webbrowser
            webbrowser.open(link)    
        gtk.about_dialog_set_url_hook(on_about_dialog_url)
        gtk.rc_parse_string('''
style "hotwire-tab-close" {
  xthickness = 0
  ythickness = 0
}
widget "*hotwire-tab-close" style "hotwire-tab-close"
''')    
        if os.getenv('HOTWIRE_UNINSTALLED'):
            theme = gtk.icon_theme_get_default()
            imgpath = os.path.join(os.getenv('HOTWIRE_UNINSTALLED'), 'images')
            _logger.debug("appending to icon theme: %s", imgpath)
            theme.append_search_path(imgpath)    
    
        factory = app.get_factory()
        factory.connect('shutdown', app.on_shutdown)
        args = sys.argv[1:]
        if len(args) == 0:
            app.offer_load_session()
        else:
            w = factory.create_initial_window()
            w.new_tab(args, os.getcwd())
            w.show_all()
            w.present()
    
        uiproxy = remote.get_proxy(factory)    

        _logger.debug('entering mainloop')
        gtk.gdk.threads_enter()
        gtk.main()
        gtk.gdk.threads_leave()    



#! /usr/bin/env python
# encoding: utf-8

import os,sys
import Configure,Options,Utils
import ccroot,ar
from Configure import conftest
def find_icc(conf):
	if sys.platform=='win32':
		conf.find_icc_win32()
	else:
		conf.find_icc_other()
def find_icc_win32(conf):
	v=conf.env
	cc=None
	if v['CC']:cc=v['CC']
	elif'CC'in os.environ:cc=os.environ['CC']
	if not cc:cc=conf.find_program('ICL',var='CC')
	if not cc:conf.fatal('Intel C Compiler (icl.exe) was not found')
	v['CC']=Utils.quote_whitespace(cc)
	v['CC_NAME']='icc'
	v['CXX']=v['CC']
def find_icc_other(conf):
	v=conf.env
	cc=None
	if v['CC']:cc=v['CC']
	elif'CC'in os.environ:cc=os.environ['CC']
	if not cc:cc=conf.find_program('icc',var='CC')
	if not cc:conf.fatal('Intel C Compiler (icc) was not found')
	try:
		if Utils.cmd_output('%s --version'%cc).find('icc')<0:
			conf.fatal('icc was not found, see the result of icc --version')
	except ValueError:
		conf.fatal('icc --version could not be executed')
	v['CC']=cc
	v['CC_NAME']='icc'
	ccroot.get_cc_version(conf,cc,'CC_VERSION')
def icc_modifier_win32(conf):
	if sys.platform!='win32':return
	v=conf.env
	try:
		v['CC_NAME']='icc'
		v['CPPFLAGS'].remove('/errorReport:prompt')
	except ValueError:
		pass
if sys.platform=='win32':
	detect='''
find_icc
find_msvc
msvc_common_flags
icc_modifier_win32
cc_load_tools
cc_add_flags
'''
elif sys.platform=='cygwin':
	conf.fatal("ICC module is not supported for cygwin. Try under native Win32")
else:
	detect='''
find_icc
find_ar
gcc_common_flags
gcc_modifier_darwin
cc_load_tools
cc_add_flags
'''

conftest(find_icc)
conftest(find_icc_win32)
conftest(find_icc_other)
conftest(icc_modifier_win32)



# This is lifted from Django:
# http://code.djangoproject.com/browser/django/trunk/django/utils/timesince.py
#Copyright (c) 2005, the Lawrence Journal-World
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification,
#are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice, 
#       this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright 
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of Django nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
#ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
from gettext import ngettext

def timesince(d, now=None):
    """
    Takes two datetime objects and returns the time between d and now
    as a nicely formatted string, e.g. "10 minutes".  If d occurs after now,
    then "0 minutes" is returned.

    Units used are years, months, weeks, days, hours, and minutes.
    Seconds and microseconds are ignored.  Up to two adjacent units will be
    displayed.  For example, "2 weeks, 3 days" and "1 year, 3 months" are
    possible outputs, but "2 weeks, 3 hours" and "1 year, 5 days" are not.

    Adapted from http://blog.natbat.co.uk/archive/2003/Jun/14/time_since
    """
    chunks = (
      (60 * 60 * 24 * 365, lambda n: ngettext('year', 'years', n)),
      (60 * 60 * 24 * 30, lambda n: ngettext('month', 'months', n)),
      (60 * 60 * 24 * 7, lambda n : ngettext('week', 'weeks', n)),
      (60 * 60 * 24, lambda n : ngettext('day', 'days', n)),
      (60 * 60, lambda n: ngettext('hour', 'hours', n)),
      (60, lambda n: ngettext('minute', 'minutes', n))
    )
    # Convert datetime.date to datetime.datetime for comparison
    if d.__class__ is not datetime.datetime:
        d = datetime.datetime(d.year, d.month, d.day)
    if now:
        t = now.timetuple()
    else:
        t = time.localtime()
    tz = None
    now = datetime.datetime(t[0], t[1], t[2], t[3], t[4], t[5], tzinfo=tz)

    # ignore microsecond part of 'd' since we removed it from 'now'
    delta = now - (d - datetime.timedelta(0, 0, d.microsecond))
    since = delta.days * 24 * 60 * 60 + delta.seconds
    if since <= 0:
        # d is in the future compared to now, stop processing.
        return _('in the future')
    if since < 60:
        return _('less than 1 minute ago')    
    for i, (seconds, name) in enumerate(chunks):
        count = since // seconds
        if count != 0:
            break
    s = None   
    if i + 1 < len(chunks):
        # Now get the second item
        seconds2, name2 = chunks[i + 1]
        count2 = (since - (seconds * count)) // seconds2
        if count2 != 0:        
            s = _('%(number)d %(type)s, %(number2)d %(type2)s ago') % {'number': count, 'type': name(count), 
                                                                       'number2': count2, 'type2': name2(count2)}        
    if not s:
        s = _('%(number)d %(type)s ago') % {'number': count, 'type': name(count)}        
    return s



#! /usr/bin/env python
# encoding: utf-8

import os,sys,imp,types,tempfile
from optparse import OptionParser
import Logs,Utils
from Constants import*
cmds='dist configure clean distclean build install uninstall check distcheck'.split()
options={}
commands={}
launch_dir=''
tooldir=''
lockfile=os.environ.get('WAFLOCK','.lock-wscript')
try:cache_global=os.path.abspath(os.environ['WAFCACHE'])
except KeyError:cache_global=''
platform=Utils.detect_platform()
conf_file='conf-runs-%s-%d.pickle'%(platform,ABI)
is_install=False
default_prefix=os.environ.get('PREFIX')
if not default_prefix:
	if platform=='win32':default_prefix=tempfile.gettempdir()
	else:default_prefix='/usr/local/'
default_jobs=os.environ.get('JOBS',-1)
if default_jobs<1:
	try:
		default_jobs=os.sysconf('SC_NPROCESSORS_ONLN')
	except:
		default_jobs=int(os.environ.get('NUMBER_OF_PROCESSORS',1))
default_destdir=os.environ.get('DESTDIR','')
def create_parser():
	Logs.debug('options: create_parser is called')
	parser=OptionParser(conflict_handler="resolve",usage="""waf [options] [commands ...]

* Main commands: configure build install clean dist distclean uninstall distcheck
* Example: ./waf build -j4""",version='waf %s'%WAFVERSION)
	parser.formatter.width=Utils.get_term_cols()
	p=parser.add_option
	p('-j','--jobs',type='int',default=default_jobs,help="amount of parallel jobs [Default: %s]"%default_jobs,dest='jobs')
	p('-f','--force',action='store_true',default=False,help='force file installation',dest='force')
	p('-k','--keep',action='store_true',default=False,help='keep running happily on independent task groups',dest='keep')
	p('-p','--progress',action='count',default=0,help='-p: progress bar; -pp: ide output',dest='progress_bar')
	p('-v','--verbose',action='count',default=0,help='verbosity level -v -vv or -vvv [Default: 0]',dest='verbose')
	p('--destdir',help="installation root [Default: '%s']"%default_destdir,default=default_destdir,dest='destdir')
	p('--nocache',action='store_true',default=False,help='compile everything, even if WAFCACHE is set',dest='nocache')
	if'configure'in sys.argv:
		p('-b','--blddir',action='store',default='',help='build dir for the project (configuration)',dest='blddir')
		p('-s','--srcdir',action='store',default='',help='src dir for the project (configuration)',dest='srcdir')
		p('--prefix',help="installation prefix (configuration only) [Default: '%s']"%default_prefix,default=default_prefix,dest='prefix')
	p('--zones',action='store',default='',help='debugging zones (task_gen, deps, tasks, etc)',dest='zones')
	p('--targets',action='store',default='',help='compile the targets given only [targets in CSV format, e.g. "target1,target2"]',dest='compile_targets')
	return parser
def parse_args_impl(parser,_args=None):
	global options,commands
	(options,args)=parser.parse_args(args=_args)
	commands={}
	for var in cmds:commands[var]=0
	if len(args)==0:commands['build']=1
	for arg in args:
		arg=arg.strip()
		if arg in cmds:
			commands[arg]=True
		else:
			print'Error: Invalid command specified ',arg
			parser.print_help()
			sys.exit(1)
	if commands['check']:
		commands['build']=True
	if commands['install']or commands['uninstall']:
		global is_install
		is_install=True
	if options.keep:options.jobs=1
	if options.jobs<1:options.jobs=1
	Logs.verbose=options.verbose
	Logs.init_log()
	if options.zones:
		Logs.zones=options.zones.split(',')
		if not Logs.verbose:Logs.verbose=1
	elif Logs.verbose==1:
		Logs.zones=['runner']
class Handler(object):
	parser=None
	def __init__(self):
		self.parser=create_parser()
		self.cwd=os.getcwd()
		Handler.parser=self
	def add_option(self,*kw,**kwargs):
		self.parser.add_option(*kw,**kwargs)
	def add_option_group(self,*args,**kwargs):
		return self.parser.add_option_group(*args,**kwargs)
	def get_option_group(self,opt_str):
		return self.parser.get_option_group(opt_str)
	def sub_options(self,dir,option_group=None):
		try:
			current=self.cwd
			self.cwd=os.path.join(self.cwd,dir)
			cur=os.path.join(self.cwd,WSCRIPT_FILE)
			mod=Utils.load_module(cur)
			if not hasattr(mod,'set_options'):
				msg="the module %s has no set_options function;\n* make sure such a function is defined\n* run configure from the root of the project"%cur
				raise Utils.WscriptError(msg)
			else:
				fun=mod.set_options
			fun(option_group or self)
		finally:
			self.cwd=current
	def tool_options(self,tool,tdir=None,option_group=None):
		
		if type(tool)is types.ListType:
			for i in tool:self.tool_options(i,tdir,option_group)
			return
		if not tdir:tdir=tooldir
		tdir=Utils.to_list(tdir)
		try:
			file,name,desc=imp.find_module(tool,tdir)
		except ImportError:
			raise Utils.WscriptError("no tool named '%s' found"%tool)
		module=imp.load_module(tool,file,name,desc)
		try:
			fun=module.set_options
		except AttributeError:
			pass
		else:
			fun(option_group or self)
	def parse_args(self,args=None):
		parse_args_impl(self.parser,args)




#! /usr/bin/env python
# encoding: utf-8

import os,re
import TaskGen,Task,Utils,Runner,Options,Build
import cc
from Logs import error
class intltool_in_taskgen(TaskGen.task_gen):
	def __init__(self,*k,**kw):
		TaskGen.task_gen.__init__(self,*k,**kw)
		self.source=''
		self.flags=''
		self.podir='po'
		self.intlcache='.intlcache'
		self.tasks=[]
	def apply(self):
		self.env=self.env.copy()
		tree=Build.bld
		for i in self.to_list(self.source):
			node=self.path.find_resource(i)
			podirnode=self.path.find_dir(self.podir)
			self.env['INTLCACHE']=os.path.join(self.path.bldpath(self.env),self.podir,self.intlcache)
			self.env['INTLPODIR']=podirnode.srcpath(self.env)
			self.env['INTLFLAGS']=self.flags
			task=self.create_task('intltool')
			task.set_inputs(node)
			task.set_outputs(node.change_ext(''))
			task.install_path=self.install_path
class intltool_po_taskgen(TaskGen.task_gen):
	def __init__(self,*k,**kw):
		TaskGen.task_gen.__init__(self,*k,**kw)
		self.default_install_path='${LOCALEDIR}'
		self.appname=kw.get('appname','set_your_app_name')
		self.podir=''
		self.tasks=[]
	def apply(self):
		def install_translation(task):
			out=task.outputs[0]
			filename=out.name
			(langname,ext)=os.path.splitext(filename)
			inst_file=langname+os.sep+'LC_MESSAGES'+os.sep+self.appname+'.mo'
			Build.bld.install_as(os.path.join(self.install_path,inst_file),out.abspath(self.env),chmod=self.chmod)
		linguas=self.path.find_resource(os.path.join(self.podir,'LINGUAS'))
		if linguas:
			file=open(linguas.abspath())
			langs=[]
			for line in file.readlines():
				if not line.startswith('#'):
					langs+=line.split()
			file.close()
			re_linguas=re.compile('[-a-zA-Z_@.]+')
			for lang in langs:
				if re_linguas.match(lang):
					node=self.path.find_resource(os.path.join(self.podir,re_linguas.match(lang).group()+'.po'))
					task=self.create_task('po')
					task.set_inputs(node)
					task.set_outputs(node.change_ext('.mo'))
					if Options.is_install:task.install=install_translation
		else:
			Utils.pprint('RED',"Error no LINGUAS file found in po directory")
Task.simple_task_type('po','${POCOM} -o ${TGT} ${SRC}',color='BLUE')
Task.simple_task_type('intltool','${INTLTOOL} ${INTLFLAGS} -q -u -c ${INTLCACHE} ${INTLPODIR} ${SRC} ${TGT}',color='BLUE',after="cc_link cxx_link")
def detect(conf):
	pocom=conf.find_program('msgfmt')
	if not pocom:
		conf.fatal('The program msgfmt (gettext) is mandatory!')
	conf.env['POCOM']=pocom
	intltool=conf.find_program('intltool-merge')
	if not intltool:
		conf.fatal('The program intltool-merge (intltool, gettext-devel) is mandatory!')
	conf.env['INTLTOOL']=intltool
	def getstr(varname):
		return getattr(Options.options,varname,'')
	prefix=conf.env['PREFIX']
	datadir=getstr('datadir')
	if not datadir:datadir=os.path.join(prefix,'share')
	conf.define('LOCALEDIR',os.path.join(datadir,'locale'))
	conf.define('DATADIR',datadir)
	conf.check(header_name='locale.h')
def set_options(opt):
	opt.add_option('--want-rpath',type='int',default=1,dest='want_rpath',help='set rpath to 1 or 0 [Default 1]')
	opt.add_option('--datadir',type='string',default='',dest='datadir',help='read-only application data')




# This file is part of the Hotwire Shell project API.

# Copyright (C) 2007 Colin Walters <walters@verbum.org>

# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal 
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
# of the Software, and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all 
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
# PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE X CONSORTIUM BE 
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR 
# THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__version__ = '0.201'

def svn_version_str():
    if not svn_version_info:
        return ''
    return '(%s  %s)' % (svn_version_info['Revision'], svn_version_info['Last Changed Date'])

# Default to empty
svn_version_info = None



#! /usr/bin/env python
# encoding: utf-8

import os,sys,types
import Utils,Build
UNDEFINED=0
DIR=1
FILE=2
BUILD=3
type_to_string={UNDEFINED:"unk",DIR:"dir",FILE:"src",BUILD:"bld"}
class Node(object):
	__slots__=("name","parent","id","childs")
	def __init__(self,name,parent,node_type=UNDEFINED):
		self.name=name
		self.parent=parent
		self.__class__.bld.id_nodes+=4
		self.id=self.__class__.bld.id_nodes+node_type
		if node_type==DIR:self.childs={}
		if Utils.split_path(name)[0]!=name:
			raise Utils.WafError('name forbidden '+name)
		if parent and name in parent.childs:
			raise Utils.WafError('node %s exists in the parent files %s already'%(name,str(parent)))
		if parent:parent.childs[name]=self
	def __str__(self):
		if not self.parent:return''
		return"%s://%s"%(type_to_string[self.id&3],self.abspath())
	def __repr__(self):
		return self.__str__()
	def __hash__(self):
		raise Utils.WafError('nodes, you are doing it wrong')
	def get_type(self):
		return self.id&3
	def set_type(self,t):
		self.id=self.id+t-self.id&3
	def dirs(self):
		return[x for x in self.childs.values()if x.id&3==DIR]
	def files(self):
		return[x for x in self.childs.values()if x.id&3==FILE]
	def get_dir(self,name,default=None):
		node=self.childs.get(name,None)
		if not node or node.id&3!=DIR:return default
		return node
	def get_file(self,name,default=None):
		node=self.childs.get(name,None)
		if not node or node.id&3!=FILE:return default
		return node
	def get_build(self,name,default=None):
		node=self.childs.get(name,None)
		if not node or node.id&3!=BUILD:return default
		return node
	def find_resource(self,lst):
		if type(lst)is types.StringType:
			lst=Utils.split_path(lst)
		if not lst[:-1]:
			parent=self
		else:
			parent=self.find_dir(lst[:-1])
			if not parent:return None
		self.__class__.bld.rescan(parent)
		name=lst[-1]
		node=parent.childs.get(name,None)
		if node:
			tp=node.id&3
			if tp==FILE or tp==BUILD:
				return node
		tree=self.__class__.bld
		if not name in tree.cache_dir_contents[parent.id]:
			return None
		path=parent.abspath()+os.sep+name
		try:
			st=Utils.h_file(path)
		except IOError:
			return None
		child=self.__class__(name,parent,FILE)
		tree.node_sigs[0][child.id]=st
		return child
	def find_or_declare(self,lst):
		if type(lst)is types.StringType:
			lst=Utils.split_path(lst)
		if not lst[:-1]:
			parent=self
		else:
			parent=self.find_dir(lst[:-1])
			if not parent:return None
		self.__class__.bld.rescan(parent)
		name=lst[-1]
		node=parent.childs.get(name,None)
		if node:
			tp=node.id&3
			if tp!=BUILD:
				raise Utils.WafError("find_or_declare returns a build node, not a source nor a directory"+str(lst))
			return node
		node=self.__class__(name,parent,BUILD)
		return node
	def find_dir(self,lst):
		if type(lst)is types.StringType:
			lst=Utils.split_path(lst)
		current=self
		for name in lst:
			self.__class__.bld.rescan(current)
			prev=current
			if not current.parent and name==current.name:
				continue
			elif not name:
				continue
			elif name=='.':
				continue
			elif name=='..':
				current=current.parent or current
			else:
				current=prev.childs.get(name,None)
				if current is None:
					dir_cont=self.__class__.bld.cache_dir_contents
					if prev.id in dir_cont and name in dir_cont[prev.id]:
						current=self.__class__(name,prev,DIR)
					else:
						return None
		return current
	def ensure_dir_node_from_path(self,lst):
		if type(lst)is types.StringType:
			lst=Utils.split_path(lst)
		current=self
		for name in lst:
			if not name:
				continue
			elif name=='.':
				continue
			elif name=='..':
				current=current.parent or current
			else:
				prev=current
				current=prev.childs.get(name,None)
				if current is None:
					current=self.__class__(name,prev,DIR)
		return current
	def exclusive_build_node(self,path):
		lst=Utils.split_path(path)
		name=lst[-1]
		if len(lst)>1:
			parent=None
			try:
				parent=self.find_dir(lst[:-1])
			except OSError:
				pass
			if not parent:
				parent=self.ensure_dir_node_from_path(lst[:-1])
				self.__class__.bld.cache_scanned_folders[parent.id]=1
			else:
				try:
					self.__class__.bld.rescan(parent)
				except OSError:
					pass
		else:
			parent=self
		node=parent.childs.get(name,None)
		if not node:
			node=self.__class__(name,parent,BUILD)
		return node
	def path_to_parent(self,parent):
		lst=[]
		p=self
		h1=parent.height()
		h2=p.height()
		while h2>h1:
			h2-=1
			lst.append(p.name)
			p=p.parent
		if lst:
			lst.reverse()
			ret=os.path.join(*lst)
		else:
			ret=''
		return ret
	def find_ancestor(self,node):
		dist=self.height()-node.height()
		if dist<0:return node.find_ancestor(self)
		cand=self
		while dist>0:
			cand=cand.parent
			dist-=1
		if cand==node:return cand
		cursor=node
		while cand.parent:
			cand=cand.parent
			cursor=cursor.parent
			if cand==cursor:return cand
	def relpath_gen(self,going_to):
		if self==going_to:return'.'
		if going_to.parent==self:return'..'
		ancestor=self.find_ancestor(going_to)
		lst=[]
		cand=self
		while not cand.id==ancestor.id:
			lst.append(cand.name)
			cand=cand.parent
		cand=going_to
		while not cand.id==ancestor.id:
			lst.append('..')
			cand=cand.parent
		lst.reverse()
		return os.sep.join(lst)
	def nice_path(self,env=None):
		tree=self.__class__.bld
		ln=tree.launch_node()
		name=self.name
		if self.id&3==FILE:return self.relpath_gen(ln)
		else:return os.path.join(tree.bldnode.relpath_gen(ln),env.variant(),self.relpath_gen(tree.srcnode))
	def is_child_of(self,node):
		p=self
		diff=self.height()-node.height()
		while diff>0:
			diff-=1
			p=p.parent
		return p.id==node.id
	def variant(self,env):
		if not env:return 0
		elif self.id&3==FILE:return 0
		else:return env.variant()
	def height(self):
		d=self
		val=-1
		while d:
			d=d.parent
			val+=1
		return val
	def abspath(self,env=None):
		if not self.name:
			return'/'
		variant=self.variant(env)
		ret=self.__class__.bld.cache_node_abspath[variant].get(self.id,None)
		if ret:return ret
		if not variant:
			if not self.parent:
				val=os.sep
			elif not self.parent.name:
				val=os.sep+self.name
			else:
				val=self.parent.abspath()+os.sep+self.name
		else:
			val=os.sep.join((self.__class__.bld.bldnode.abspath(),env.variant(),self.path_to_parent(self.__class__.bld.srcnode)))
		self.__class__.bld.cache_node_abspath[variant][self.id]=val
		return val
	def change_ext(self,ext):
		name=self.name
		k=name.rfind('.')
		if k>=0:
			name=name[:k]+ext
		else:
			name=name+ext
		node=self.parent.childs.get(name,None)
		if not node:
			node=self.__class__(name,self.parent,BUILD)
		return node
	def src_dir(self,env):
		return self.parent.srcpath(env)
	def bld_dir(self,env):
		return self.parent.bldpath(env)
	def bld_base(self,env):
		s=self.name
		s=s[:s.rfind('.')]
		return os.path.join(self.bld_dir(env),s)
	def bldpath(self,env=None):
		if self.id&3==FILE:
			return self.relpath_gen(self.__class__.bld.bldnode)
		if self.path_to_parent(self.__class__.bld.srcnode)is not'':
			return os.path.join(env.variant(),self.path_to_parent(self.__class__.bld.srcnode))
		return env.variant()
	def srcpath(self,env=None):
		if self.id&3==BUILD:
			return self.bldpath(env)
		return self.relpath_gen(self.__class__.bld.bldnode)
	def read(self,env):
		try:
			file=open(self.abspath(env),'rb')
			return file.read()
		finally:
			if file:file.close()
	def dir(self,env):
		return self.parent.abspath(env)
	def file(self):
		return self.name
	def file_base(self):
		s=self.name
		if s.rfind('.')<0:
			return s
		return s[:s.rfind('.')]
	def suffix(self):
		s=self.name
		if s.rfind('.')<0:
			return s
		return s[s.rfind('.'):]
if sys.platform=="win32":
	def find_dir_win32(self,lst):
		if type(lst)is types.StringType:
			lst=Utils.split_path(lst)
		current=self
		for name in lst:
			self.__class__.bld.rescan(current)
			prev=current
			if not current.parent and name==current.name:
				continue
			if not name:
				continue
			elif name=='.':
				continue
			elif name=='..':
				current=current.parent or current
			else:
				current=prev.childs.get(name,None)
				if current is None:
					if(name in self.__class__.bld.cache_dir_contents[prev.id]or(not prev.parent and name[1]==":")):
						current=self.__class__(name,prev,DIR)
					else:
						return None
		return current
	Node.find_dir=find_dir_win32
	def abspath_win32(self,env=None):
		variant=self.variant(env)
		ret=self.__class__.bld.cache_node_abspath[variant].get(self.id,None)
		if ret:return ret
		if not variant:
			cur=self
			lst=[]
			while cur:
				lst.append(cur.name)
				cur=cur.parent
			lst.reverse()
			val=os.sep.join(lst)
		else:
			val=os.sep.join((self.__class__.bld.bldnode.abspath(),env.variant(),self.path_to_parent(self.__class__.bld.srcnode)))
		if val.startswith("\\"):val=val[1:]
		if val.startswith("\\"):val=val[1:]
		self.__class__.bld.cache_node_abspath[variant][self.id]=val
		return val
	Node.abspath=abspath_win32
class Nodu(Node):
	pass




#! /usr/bin/env python
# encoding: utf-8

import os,sys,imp,types,ccroot
import optparse
import Utils,Configure,Options
c_compiler={'win32':['msvc','gcc'],'cygwin':['gcc'],'darwin':['gcc'],'aix5':['gcc'],'linux':['gcc','suncc'],'sunos':['suncc','gcc'],'irix':['gcc'],'hpux':['gcc'],'default':['gcc']}
def __list_possible_compiler(platform):
	try:
		return c_compiler[platform]
	except KeyError:
		return c_compiler["default"]
def detect(conf):
	try:test_for_compiler=Options.options.check_c_compiler
	except AttributeError:raise Configure.ConfigurationError("Add set_options(opt): opt.tool_options('compiler_cc')")
	for c_compiler in test_for_compiler.split():
		conf.check_tool(c_compiler)
		if conf.env['CC']:
			conf.check_message("%s"%c_compiler,'',True)
			conf.env["COMPILER_CC"]="%s"%c_compiler
			return
		conf.check_message("%s"%c_compiler,'',False)
	conf.env["COMPILER_CC"]=None
def set_options(opt):
	detected_platform=Options.platform
	possible_compiler_list=__list_possible_compiler(detected_platform)
	test_for_compiler=str(" ").join(possible_compiler_list)
	cc_compiler_opts=opt.add_option_group("C Compiler Options")
	cc_compiler_opts.add_option('--check-c-compiler',default="%s"%test_for_compiler,help='On this platform (%s) the following C-Compiler will be checked by default: "%s"'%(detected_platform,test_for_compiler),dest="check_c_compiler")
	for c_compiler in test_for_compiler.split():
		opt.tool_options('%s'%c_compiler,option_group=cc_compiler_opts)




#! /usr/bin/env python
# encoding: utf-8

import os,sys,re,string,optparse
import Utils,TaskGen,Runner,Configure,Task,Options
from Logs import debug,error,warn
from Utils import quote_whitespace
from TaskGen import taskgen,after,before,feature
from Configure import conftest
import ccroot,cc,cxx,ar
from libtool import read_la_file
from os.path import exists
def msvc_linker(task):
	e=task.env
	linker=e['LINK']
	srcf=e['LINK_SRC_F']
	trgtf=e['LINK_TGT_F']
	linkflags=e.get_flat('LINKFLAGS')
	libdirs=e.get_flat('_LIBDIRFLAGS')
	libs=e.get_flat('_LIBFLAGS')
	subsystem=''
	if task.subsystem:
		subsystem='/subsystem:%s'%task.subsystem
	outfile=task.outputs[0].bldpath(e)
	manifest=outfile+'.manifest'
	pdbnode=task.outputs[0].change_ext('.pdb')
	pdbfile=pdbnode.bldpath(e)
	objs=" ".join(['"%s"'%a.abspath(e)for a in task.inputs])
	cmd="%s %s %s%s %s%s %s %s %s"%(linker,subsystem,srcf,objs,trgtf,outfile,linkflags,libdirs,libs)
	ret=task.generator.bld.exec_command(cmd,shell=False)
	if ret:return ret
	if os.path.exists(pdbfile):
		task.outputs.append(pdbnode)
	if os.path.exists(manifest):
		debug('msvc: manifesttool')
		mtool=e['MT']
		if not mtool:
			return 0
		mode=''
		if'cprogram'in task.features:
			mode='1'
		elif'cshlib'in task.features:
			mode='2'
		debug('msvc: embedding manifest')
		flags=' '.join(e['MTFLAGS']or[])
		cmd='%s %s -manifest "%s" -outputresource:"%s";#%s'%(mtool,flags,manifest,outfile,mode)
		ret=task.generator.bld.exec_command(cmd,shell=False)
	return ret
g_msvc_systemlibs="""
aclui activeds ad1 adptif adsiid advapi32 asycfilt authz bhsupp bits bufferoverflowu cabinet
cap certadm certidl ciuuid clusapi comctl32 comdlg32 comsupp comsuppd comsuppw comsuppwd comsvcs
credui crypt32 cryptnet cryptui d3d8thk daouuid dbgeng dbghelp dciman32 ddao35 ddao35d
ddao35u ddao35ud delayimp dhcpcsvc dhcpsapi dlcapi dnsapi dsprop dsuiext dtchelp
faultrep fcachdll fci fdi framedyd framedyn gdi32 gdiplus glauxglu32 gpedit gpmuuid
gtrts32w gtrtst32hlink htmlhelp httpapi icm32 icmui imagehlp imm32 iphlpapi iprop
kernel32 ksguid ksproxy ksuser libcmt libcmtd libcpmt libcpmtd loadperf lz32 mapi
mapi32 mgmtapi minidump mmc mobsync mpr mprapi mqoa mqrt msacm32 mscms mscoree
msdasc msimg32 msrating mstask msvcmrt msvcurt msvcurtd mswsock msxml2 mtx mtxdm
netapi32 nmapinmsupp npptools ntdsapi ntdsbcli ntmsapi ntquery odbc32 odbcbcp
odbccp32 oldnames ole32 oleacc oleaut32 oledb oledlgolepro32 opends60 opengl32
osptk parser pdh penter pgobootrun pgort powrprof psapi ptrustm ptrustmd ptrustu
ptrustud qosname rasapi32 rasdlg rassapi resutils riched20 rpcndr rpcns4 rpcrt4 rtm
rtutils runtmchk scarddlg scrnsave scrnsavw secur32 sensapi setupapi sfc shell32
shfolder shlwapi sisbkup snmpapi sporder srclient sti strsafe svcguid tapi32 thunk32
traffic unicows url urlmon user32 userenv usp10 uuid uxtheme vcomp vcompd vdmdbg
version vfw32 wbemuuid  webpost wiaguid wininet winmm winscard winspool winstrm
wintrust wldap32 wmiutils wow32 ws2_32 wsnmp32 wsock32 wst wtsapi32 xaswitch xolehlp
""".split()
def find_lt_names_msvc(self,libname,is_static=False):
	lt_names=['lib%s.la'%libname,'%s.la'%libname,]
	for path in self.libpaths:
		for la in lt_names:
			laf=os.path.join(path,la)
			dll=None
			if exists(laf):
				ltdict=read_la_file(laf)
				lt_libdir=None
				if ltdict.has_key('libdir')and ltdict['libdir']!='':
					lt_libdir=ltdict['libdir']
				if not is_static and ltdict.has_key('library_names')and ltdict['library_names']!='':
					dllnames=ltdict['library_names'].split()
					dll=dllnames[0].lower()
					dll=re.sub('\.dll$','',dll)
					return(lt_libdir,dll,False)
				elif ltdict.has_key('old_library')and ltdict['old_library']!='':
					olib=ltdict['old_library']
					if exists(os.path.join(path,olib)):
						return(path,olib,True)
					elif lt_libdir!=''and exists(os.path.join(lt_libdir,olib)):
						return(lt_libdir,olib,True)
					else:
						return(None,olib,True)
				else:
					raise Utils.WafError('invalid libtool object file: %s'%laf)
	return(None,None,None)
def libname_msvc(self,libname,is_static=False):
	lib=libname.lower()
	lib=re.sub('\.lib$','',lib)
	if lib in g_msvc_systemlibs:
		return lib+'.lib'
	lib=re.sub('^lib','',lib)
	if lib=='m':
		return None
	(lt_path,lt_libname,lt_static)=find_lt_names_msvc(self,lib,is_static)
	if lt_path!=None and lt_libname!=None:
		if lt_static==True:
			return os.path.join(lt_path,lt_libname)
	if lt_path!=None:
		_libpaths=[lt_path]+self.libpaths
	else:
		_libpaths=self.libpaths
	static_libs=['%ss.lib'%lib,'lib%ss.lib'%lib,'%s.lib'%lib,'lib%s.lib'%lib,]
	dynamic_libs=['lib%s.dll.lib'%lib,'lib%s.dll.a'%lib,'%s.dll.lib'%lib,'%s.dll.a'%lib,'lib%s_d.lib'%lib,'%s_d.lib'%lib,'%s.lib'%lib,]
	libnames=static_libs
	if not is_static:
		libnames=dynamic_libs+static_libs
	for path in _libpaths:
		for libn in libnames:
			if os.path.exists(os.path.join(path,libn)):
				debug('msvc: lib found: %s'%os.path.join(path,libn))
				return libn
	return None
def apply_obj_vars(self):
	debug('msvc: apply_msvc_obj_vars called for msvc')
	env=self.env
	app=env.append_unique
	cpppath_st=env['CPPPATH_ST']
	lib_st=env['LIB_ST']
	staticlib_st=env['STATICLIB_ST']
	libpath_st=env['LIBPATH_ST']
	staticlibpath_st=env['STATICLIBPATH_ST']
	for i in env['LIBPATH']:
		app('LINKFLAGS',libpath_st%i)
		if not self.libpaths.count(i):
			self.libpaths.append(i)
	for i in env['LIBPATH']:
		app('LINKFLAGS',staticlibpath_st%i)
		if not self.libpaths.count(i):
			self.libpaths.append(i)
	if not env['FULLSTATIC']:
		if env['STATICLIB']or env['LIB']:
			app('LINKFLAGS',env['SHLIB_MARKER'])
	if env['STATICLIB']:
		app('LINKFLAGS',env['STATICLIB_MARKER'])
		for i in env['STATICLIB']:
			debug('msvc: libname: %s'%i)
			libname=libname_msvc(self,i,True)
			debug('msvc: libnamefixed: %s'%libname)
			if libname!=None:
				app('LINKFLAGS',libname)
	if self.env['LIB']:
		for i in env['LIB']:
			debug('msvc: libname: %s'%i)
			libname=libname_msvc(self,i)
			debug('msvc: libnamefixed: %s'%libname)
			if libname!=None:
				app('LINKFLAGS',libname)
def apply_link(self):
	if'objects'in self.features:
		self.out_nodes=[]
		app=self.out_nodes.append
		for t in self.compiled_tasks:app(t.outputs[0])
		return
	link=getattr(self,'link',None)
	if not link:
		if'cstaticlib'in self.features:link='msvc_ar_link_static'
		elif'cxx'in self.features:link='msvc_cxx_link'
		else:link='msvc_cc_link'
	linktask=self.create_task(link)
	outputs=[t.outputs[0]for t in self.compiled_tasks]
	linktask.set_inputs(outputs)
	linktask.set_outputs(self.path.find_or_declare(ccroot.get_target_name(self)))
	linktask.features=self.features
	linktask.subsystem=getattr(self,'subsystem','')
	self.link_task=linktask
def init_msvc(self):
	try:_libpaths=getattr(self,'libpaths')
	except AttributeError:self.libpaths=[]
static_link_str='${STLIBLINK} ${STLINKFLAGS} ${LINK_SRC_F}${SRC} ${LINK_TGT_F}${TGT}'
Task.simple_task_type('msvc_ar_link_static',static_link_str,color='YELLOW',ext_in='.o')
Task.task_type_from_func('msvc_cc_link',vars=['LINK','LINK_SRC_F','LINK_TGT_F','LINKFLAGS','_LIBDIRFLAGS','_LIBFLAGS','MT','MTFLAGS'],color='YELLOW',func=msvc_linker,ext_in='.o')
Task.task_type_from_func('msvc_cxx_link',vars=['LINK','LINK_SRC_F','LINK_TGT_F','LINKFLAGS','_LIBDIRFLAGS','_LIBFLAGS','MT','MTFLAGS'],color='YELLOW',func=msvc_linker,ext_in='.o')
rc_str='${RC} ${RCFLAGS} /fo ${TGT} ${SRC}'
Task.simple_task_type('rc',rc_str,color='GREEN',before='cc cxx')
import winres
detect='''
find_msvc
msvc_common_flags
cc_load_tools
cxx_load_tools
cc_add_flags
cxx_add_flags
'''
def find_msvc(conf):
	if sys.platform!='win32':
		conf.fatal('MSVC module only works under native Win32 Python! cygwin is not supported yet')
	v=conf.env
	cxx=None
	if v['CXX']:cxx=v['CXX']
	elif'CXX'in os.environ:cxx=os.environ['CXX']
	if not cxx:cxx=conf.find_program('CL',var='CXX')
	if not cxx:conf.fatal('CL was not found (compiler)')
	v['CXX']=cxx
	v['CC']=v['CXX']
	v['CXX_NAME']='msvc'
	v['CC_NAME']='msvc'
	if not v['LINK_CXX']:
		link=conf.find_program('LINK')
		if link:v['LINK_CXX']=link
		else:conf.fatal('LINK was not found (linker)')
	v['LINK']=link
	if not v['LINK_CC']:v['LINK_CC']=v['LINK_CXX']
	if not v['STLIBLINK']:
		stliblink=conf.find_program('LIB')
		if not stliblink:return
		v['STLIBLINK']=stliblink
		v["STLINKFLAGS"]=['/NOLOGO']
	manifesttool=conf.find_program('MT')
	if manifesttool:
		v['MT']=manifesttool
		v['MTFLAGS']=['/NOLOGO']
	conf.check_tool('winres')
	if not conf.env['WINRC']:
		warn('Resource compiler not found. Compiling resource file is disabled')
def msvc_common_flags(conf):
	v=conf.env
	v['CPPFLAGS']=['/W3','/nologo','/EHsc','/errorReport:prompt']
	v['CCDEFINES']=['WIN32']
	v['CXXDEFINES']=['WIN32']
	v['_CCINCFLAGS']=[]
	v['_CCDEFFLAGS']=[]
	v['_CXXINCFLAGS']=[]
	v['_CXXDEFFLAGS']=[]
	v['CC_SRC_F']=''
	v['CC_TGT_F']='/c /Fo'
	v['CXX_SRC_F']=''
	v['CXX_TGT_F']='/c /Fo'
	v['CPPPATH_ST']='/I%s'
	v['CPPFLAGS_CONSOLE']=['/SUBSYSTEM:CONSOLE']
	v['CPPFLAGS_NATIVE']=['/SUBSYSTEM:NATIVE']
	v['CPPFLAGS_POSIX']=['/SUBSYSTEM:POSIX']
	v['CPPFLAGS_WINDOWS']=['/SUBSYSTEM:WINDOWS']
	v['CPPFLAGS_WINDOWSCE']=['/SUBSYSTEM:WINDOWSCE']
	v['CPPFLAGS_CRT_MULTITHREADED']=['/MT']
	v['CPPFLAGS_CRT_MULTITHREADED_DLL']=['/MD']
	v['CPPDEFINES_CRT_MULTITHREADED']=['_MT']
	v['CPPDEFINES_CRT_MULTITHREADED_DLL']=['_MT','_DLL']
	v['CPPFLAGS_CRT_MULTITHREADED_DBG']=['/MTd']
	v['CPPFLAGS_CRT_MULTITHREADED_DLL_DBG']=['/MDd']
	v['CPPDEFINES_CRT_MULTITHREADED_DBG']=['_DEBUG','_MT']
	v['CPPDEFINES_CRT_MULTITHREADED_DLL_DBG']=['_DEBUG','_MT','_DLL']
	v['CCFLAGS']=['/TC']
	v['CCFLAGS_OPTIMIZED']=['/O2','/DNDEBUG']
	v['CCFLAGS_RELEASE']=['/O2','/DNDEBUG']
	v['CCFLAGS_DEBUG']=['/Od','/RTC1','/D_DEBUG','/ZI']
	v['CCFLAGS_ULTRADEBUG']=['/Od','/RTC1','/D_DEBUG','/ZI']
	v['CXXFLAGS']=['/TP']
	v['CXXFLAGS_OPTIMIZED']=['/O2','/DNDEBUG']
	v['CXXFLAGS_RELEASE']=['/O2','/DNDEBUG']
	v['CXXFLAGS_DEBUG']=['/Od','/RTC1','/D_DEBUG','/ZI']
	v['CXXFLAGS_ULTRADEBUG']=['/Od','/RTC1','/D_DEBUG','/ZI']
	v['LIB']=[]
	v['LINK_TGT_F']='/OUT:'
	v['LINK_SRC_F']=' '
	v['LIB_ST']='%s.lib'
	v['LIBPATH_ST']='/LIBPATH:%s'
	v['STATICLIB_ST']='%s.lib'
	v['STATICLIBPATH_ST']='/LIBPATH:%s'
	v['CCDEFINES_ST']='/D%s'
	v['CXXDEFINES_ST']='/D%s'
	v['_LIBDIRFLAGS']=''
	v['_LIBFLAGS']=''
	v['SHLIB_MARKER']=''
	v['STATICLIB_MARKER']=''
	v['LINKFLAGS']=['/NOLOGO','/MACHINE:X86','/ERRORREPORT:PROMPT']
	try:
		debug_level=Options.options.debug_level.upper()
	except AttributeError:
		debug_level=ccroot.DEBUG_LEVELS.CUSTOM
	v['CCFLAGS']+=v['CCFLAGS_'+debug_level]
	v['CXXFLAGS']+=v['CXXFLAGS_'+debug_level]
	v['LINKFLAGS']+=v['LINKFLAGS_'+debug_level]
	v['shlib_CCFLAGS']=['']
	v['shlib_CXXFLAGS']=['']
	v['shlib_LINKFLAGS']=['/DLL']
	v['shlib_PATTERN']='%s.dll'
	v['staticlib_LINKFLAGS']=['']
	v['staticlib_PATTERN']='%s.lib'
	v['program_PATTERN']='%s.exe'
def set_options(opt):
	opt.add_option('-d','--debug-level',action='store',default=ccroot.DEBUG_LEVELS.DEBUG,help="Specify the debug level, does nothing if CFLAGS is set in the environment. [Allowed Values: '%s']"%"', '".join(ccroot.DEBUG_LEVELS.ALL),choices=ccroot.DEBUG_LEVELS.ALL,dest='debug_level')

taskgen(apply_obj_vars)
feature('cprogram','cshlib','cstaticlib')(apply_obj_vars)
after('apply_lib_vars')(apply_obj_vars)
taskgen(apply_link)
feature('cprogram','cshlib','cstaticlib')(apply_link)
after('apply_core')(apply_link)
taskgen(init_msvc)
feature('cc','cxx')(init_msvc)
after('init_cc')(init_msvc)
after('init_cxx')(init_msvc)
before('apply_type_vars')(init_msvc)
before('apply_core')(init_msvc)
conftest(find_msvc)
conftest(msvc_common_flags)



#! /usr/bin/env python
# encoding: utf-8

import os,sys,re
import TaskGen,Task
from Utils import quote_whitespace
from TaskGen import extension
EXT_WINRC=['.rc']
winrc_str='${WINRC} ${_CPPDEFFLAGS} ${_CXXDEFFLAGS} ${_CCDEFFLAGS} ${WINRCFLAGS} ${_CPPINCFLAGS} ${_CXXINCFLAGS} ${_CCINCFLAGS} ${WINRC_TGT_F}${TGT} ${WINRC_SRC_F}${SRC}'
def rc_file(self,node):
	obj_ext='.rc.o'
	if self.env['WINRC_TGT_F']=='/fo ':obj_ext='.res'
	rctask=self.create_task('winrc')
	rctask.set_inputs(node)
	rctask.set_outputs(node.change_ext(obj_ext))
	self.compiled_tasks.append(rctask)
Task.simple_task_type('winrc',winrc_str,color='BLUE',before='cc cxx')
def detect(conf):
	v=conf.env
	cc=os.path.basename(''.join(v['CC']).lower())
	cxx=os.path.basename(''.join(v['CXX']).lower())
	cc=re.sub('"','',cc)
	cxx=re.sub('"','',cxx)
	if cc in['gcc','cc','g++','c++']:
		winrc=conf.find_program('windres',var='WINRC')
		v['WINRC_TGT_F']='-o '
		v['WINRC_SRC_F']='-i '
	elif cc.startswith('cl.exe')or cxx.startswith('cl.exe'):
		winrc=conf.find_program('RC',var='WINRC')
		v['WINRC_TGT_F']='/fo '
		v['WINRC_SRC_F']=' '
	else:
		return 0
	if not winrc:
		conf.fatal('winrc was not found!!')
	else:
		v['WINRC']=quote_whitespace(winrc)
	v['WINRCFLAGS']=''

extension(EXT_WINRC)(rc_file)
